"""Module for handling tilemaps."""
from dataclasses import dataclass
from typing import Generator

import pyxel

from bansoko.graphics import Rect, Direction, Point, Layer, TILE_SIZE


@dataclass(frozen=True)
class TilePosition:
    """Position of tile in tilemap space."""
    tile_x: int = 0
    tile_y: int = 0

    def move(self, direction: Direction) -> "TilePosition":
        """Create tile position that is a result of moving this tile position in specified
        direction by one tile.

        :param direction: direction to move tile position in
        :return: newly created tile position
        """
        return TilePosition(self.tile_x + direction.dx, self.tile_y + direction.dy)

    def to_point(self) -> Point:
        """Convert tile position to a point in screen space."""
        return Point(self.tile_x * TILE_SIZE, self.tile_y * TILE_SIZE)


@dataclass(frozen=True)
class Tilemap:
    """Tilemap is a rectangular fragment of Pyxel's mega-tilemap and describes the level layout.

    Tilemaps can have multiple layers (to achieve pseudo 3d effect). In that case all layers are
    stored on separate Pyxel's mega-tilemaps. For example: tilemap with 3 layers is defined across
    Pyxel's mega-tilemaps: tilemap_id, tilemap_id+1 and tilemap_id+2.

    Attributes:
        tilemap_id - Pyxel's mega-tilemap id
        rect_uv    - coordinates of the tilemap in Pyxel's mega-tilemap (Tilemap is just a fragment
                     of Pyxel's mega-tilemap)
        num_layers - number of layers the tilemap consists of (each layer is stored in a separate
                     Pyxel's mega-tilemap)
    """
    tilemap_id: int
    rect_uv: Rect
    num_layers: int = 1

    @property
    def width(self) -> int:
        """Width of the tilemap (expressed as a number of horizontal tiles)."""
        return self.rect_uv.w

    @property
    def height(self) -> int:
        """Height of the tilemap (expressed as a number of vertical tiles)."""
        return self.rect_uv.h

    def tile_index_at(self, position: TilePosition) -> int:
        """Return index of tile at specified position.

        :param position: tile position to retrieve tile index at
        :return: index of tile at given position
        """
        tile_index: int = pyxel.tilemap(self.tilemap_id).get(
            self.rect_uv.x + position.tile_x, self.rect_uv.y + position.tile_y)
        return tile_index

    def tiles_positions(self) -> Generator[TilePosition, None, None]:
        """Generator for iterating over all valid tile positions for this tilemap."""
        for i in range(self.width * self.height):
            yield TilePosition(i % self.width, i // self.width)

    def draw(self, layer: Layer) -> None:
        """Draw tilemap on given layer.

        :param layer: layer to draw tilemap on
        """
        if layer and layer.layer_index >= self.num_layers:
            return

        pyxel.bltm(layer.offset.x, layer.offset.y, self.tilemap_id + layer.layer_index,
                   self.rect_uv.x, self.rect_uv.y, self.rect_uv.w, self.rect_uv.h,
                   colkey=layer.transparency_color)



"""Module defining game screen which is displayed after finishing the last level."""

from bansoko.game.screens.screen_factory import ScreenFactory
from bansoko.gui.menu import MenuController, TextMenuItem, Menu, MenuLayout


class VictoryController(MenuController):
    """Screen controller for displaying victory screen.

    It's displayed when player completes the last level (finishes the game).
    Player can only go back to main menu screen from here.
    """

    def __init__(self, screen_factory: ScreenFactory):
        screen = screen_factory.get_bundle().get_screen("victory")
        menu = Menu.with_defaults(tuple([
            TextMenuItem("MAIN MENU", screen_factory.get_main_menu)
        ]), MenuLayout(position=screen.menu_position))
        super().__init__(menu=menu, screen=screen)



"""Module for handling critical game errors."""
import textwrap

import pyxel

from bansoko import GAME_FRAME_TIME_IN_MS
from bansoko.graphics import Point, center_in_rect, Rect
from bansoko.graphics.text import draw_text, text_size
from bansoko.gui.menu import MenuController, Menu, TextMenuItem, MenuLayout
from bansoko.gui.navigator import ScreenNavigator

PADDING = Point(8, 8)


class ErrorScreen(MenuController):
    """Screen controller for displaying game critical errors.

    It's displayed when a non-recoverable error occurs during game startup.
    """

    def __init__(self, message: str):
        self.error_message = self._build_error_message(message)
        self.frame_rect = self._get_frame_rect(self.error_message)
        menu = Menu.with_defaults(tuple([
            TextMenuItem("OK", lambda: None)
        ]), MenuLayout(position=self._get_menu_position(self.frame_rect)))
        super().__init__(menu=menu)

    def draw(self, draw_as_secondary: bool = False) -> None:
        self._draw_background()
        self._draw_frame()
        super().draw(draw_as_secondary=draw_as_secondary)

    def _draw_frame(self) -> None:
        pyxel.rectb(self.frame_rect.x, self.frame_rect.y, self.frame_rect.w, self.frame_rect.h, 8)
        pyxel.rect(self.frame_rect.x + 1, self.frame_rect.y + 1, self.frame_rect.w - 2,
                   self.frame_rect.h - 2, 2)
        draw_text(self.frame_rect.position.offset(PADDING), self.error_message)

    @staticmethod
    def _draw_background() -> None:
        for i in range(7, 255, 16):
            pyxel.line(i, 0, i, 255, 1)
            pyxel.line(0, i, 255, i, 1)

    @staticmethod
    def _build_error_message(message: str) -> str:
        wrapped_message = "\n".join(textwrap.wrap(message, 50))
        return f"#E** ERROR **#7\n\n{wrapped_message}\n\nPlease check logs for more details."

    @staticmethod
    def _get_frame_rect(message: str) -> Rect:
        return center_in_rect(text_size(message).enlarge(
            2 * PADDING.x, 2 * PADDING.y + 3 * pyxel.FONT_HEIGHT))

    @staticmethod
    def _get_menu_position(frame_rect: Rect) -> Point:
        return Point(-1, frame_rect.bottom - PADDING.y - pyxel.FONT_HEIGHT)


def show_error_message(message: str) -> None:
    """Display an error screen with a message from which player can only quit the game.

    :param message: error message to be displayed
    """
    navigator = ScreenNavigator(ErrorScreen(message), pyxel.quit, GAME_FRAME_TIME_IN_MS)
    pyxel.run(navigator.update, navigator.draw)



"""Module exposing level template."""
from dataclasses import dataclass
from typing import Tuple, Dict

from bansoko import GAME_FRAME_TIME_IN_MS, LEVEL_WIDTH, LEVEL_HEIGHT, LEVEL_NUM_LAYERS, \
    LEVEL_BASE_TILEMAP
from bansoko.game import GameError
from bansoko.game.game_object import Crate, Robot, RobotState, CrateState
from bansoko.game.tiles import Tileset, TileType
from bansoko.graphics import Layer, Point, Rect, Direction, TILE_SIZE
from bansoko.graphics.animation import Animation
from bansoko.graphics.sprite import SpritePack, Sprite
from bansoko.graphics.tilemap import Tilemap, TilePosition


@dataclass(frozen=True)
class LevelSpritePacks:
    """
    Collection of sprite packs used by a level.

    Attributes:
        robot_sprite_pack - sprite pack containing all robot related sprites
        crate_sprite_pack - sprite pack containing all crate related sprites
    """
    robot_sprite_pack: SpritePack
    crate_sprite_pack: SpritePack

    @property
    def robot_animations(self) -> Dict[RobotState, Animation]:
        """The pack of animations for all robot states."""
        return {
            RobotState.STANDING: Animation(
                sprite=self.robot_sprite_pack.sprites[RobotState.STANDING],
                frame_length=GAME_FRAME_TIME_IN_MS),
            RobotState.MOVING: Animation(
                sprite=self.robot_sprite_pack.sprites[RobotState.MOVING],
                frame_length=GAME_FRAME_TIME_IN_MS,
                looped=True),
            RobotState.PUSHING: Animation(
                sprite=self.robot_sprite_pack.sprites[RobotState.PUSHING],
                frame_length=GAME_FRAME_TIME_IN_MS,
                looped=True)
        }

    @property
    def crate_sprites(self) -> Dict[CrateState, Sprite]:
        """The pack of sprites for all crate states."""
        return {
            CrateState.MISPLACED: self.crate_sprite_pack.sprites[0],
            CrateState.PLACED: self.crate_sprite_pack.sprites[1]
        }


@dataclass(frozen=True)
class LevelTemplate:
    """LevelTemplate is a blue-print used for level creation.

    Attributes:
        level_num - the level number
        tilemap - tilemap to be used in the level
        tileset -  tileset to be used in the level
        layers - list of layers level will be drawn on
        sprite_packs - sprite packs to be used in the level
    """
    level_num: int
    tilemap: Tilemap
    tileset: Tileset
    layers: Tuple[Layer, ...]
    sprite_packs: LevelSpritePacks

    @classmethod
    def from_level_num(cls, level_num: int, tileset_index: int, draw_offset: Point,
                       sprite_packs: LevelSpritePacks) -> "LevelTemplate":
        """Create a new level template for given level number.

        :param level_num: level number to create level template for
        :param tileset_index: index of first tile (starting tile) used in the template
        :param draw_offset: the initial offset of the level (u3sed when level is drawn)
        :param sprite_packs: sprite packs used in the level
        :return: newly created level template
        """
        tilemap_u = LEVEL_WIDTH * (level_num % TILE_SIZE)
        tilemap_v = LEVEL_HEIGHT * (level_num // TILE_SIZE)
        tilemap_uv_rect = Rect.from_coords(tilemap_u, tilemap_v, LEVEL_WIDTH, LEVEL_HEIGHT)
        tilemap = Tilemap(LEVEL_BASE_TILEMAP, tilemap_uv_rect, LEVEL_NUM_LAYERS)
        tileset = Tileset(tileset_index)
        layers = tuple(
            [Layer(i, opaque=(i == 0), global_offset=draw_offset) for i in range(LEVEL_NUM_LAYERS)])
        return cls(level_num=level_num, tilemap=tilemap, tileset=tileset, layers=layers,
                   sprite_packs=sprite_packs)

    def tile_at(self, position: TilePosition) -> TileType:
        """Return the type of tile at given position in tilemap.

        :param position: position of the tile to check tile type of
        :return: type of the tile at given position
        """
        return self.tileset.tile_of(self.tilemap.tile_index_at(position))

    def create_crates(self) -> Tuple[Crate, ...]:
        """Create a collection of crates based on information from tilemap about their initial
        positions.

        :return: collection of crates created from level template
        """
        crates_positions = []
        for tile_position in self.tilemap.tiles_positions():
            if self.tile_at(tile_position).is_crate_spawn_point:
                crates_positions.append(tile_position)

        crates = []
        for crate_position in crates_positions:
            is_initially_placed = self.tile_at(crate_position).is_crate_initially_placed
            crates.append(
                Crate(crate_position, is_initially_placed, self.sprite_packs.crate_sprites))
        if not crates:
            raise GameError(f"Level {self.level_num} does not have any crates")

        return tuple(crates)

    def create_robot(self) -> Robot:
        """Create robot instance based on information from tilemap about its start position."""
        start = None
        for tile_position in self.tilemap.tiles_positions():
            if self.tile_at(tile_position).is_start:
                start = tile_position
        if not start:
            raise GameError(f"Level {self.level_num} does not have player start tile")

        face_direction = Direction.UP
        for direction in list(Direction):
            if self.tile_at(start.move(direction)).is_walkable:
                face_direction = direction
                break

        return Robot(start, face_direction, self.sprite_packs.robot_animations)



"""Module exposing text drawing related routines."""
from dataclasses import dataclass
from typing import Optional

import pyxel

from bansoko.graphics import Size, Point


@dataclass(frozen=True)
class TextStyle:
    """TextStyle defines basic properties used when text is drawn."""
    color: int = 7
    shadow_color: Optional[int] = None
    vertical_space: int = 0


def text_size(text: str, style: TextStyle = TextStyle()) -> Size:
    """Calculate the size of given text.

    Size represents the amount of screen space occupied by text during drawing.

    :param text: string to calculate size for
    :param style: text style calculation should be performed for
    :return: size of the string that will be used during text drawing
    """
    text_width = 0
    text_height = pyxel.FONT_HEIGHT
    line_width = 0
    color_tag = False
    for char in text:
        if char == '\n':
            text_width = max(text_width, line_width)
            text_height += pyxel.FONT_HEIGHT + style.vertical_space
            line_width = 0
            color_tag = False
            continue
        if char == '#':
            color_tag = not color_tag
            if color_tag:
                continue

        if color_tag:
            color_tag = False
        else:
            line_width += pyxel.FONT_WIDTH
    text_height -= style.vertical_space
    text_width = max(text_width, line_width)
    return Size(text_width, text_height)


def draw_text(position: Point, text: str, style: TextStyle = TextStyle()) -> None:
    """Draws given text at specified position with defined text style.

    :param position: position of text to be drawn at
    :param text: text to be drawn
    :param style: style of text to be drawn with
    """
    x = position.x
    y = position.y
    current_color = style.color
    color_tag = False
    for char in text:
        if char == '\n':
            x = position.x
            y += pyxel.FONT_HEIGHT + style.vertical_space
            color_tag = False
            continue
        if char == '#':
            color_tag = not color_tag
            if color_tag:
                continue

        if color_tag:
            current_color = int(char, 16)
            color_tag = False
        else:
            if style.shadow_color is not None:
                pyxel.text(x + 1, y + 1, char, style.shadow_color)
            pyxel.text(x, y, char, current_color)
            x += pyxel.FONT_WIDTH



"""Module providing an abstraction over screen controllers creation."""
from abc import ABC, abstractmethod
from typing import Callable

from bansoko.game.bundle import Bundle
from bansoko.game.profile import PlayerProfile, LevelScore
from bansoko.gui.navigator import ScreenController


class ScreenFactory(ABC):
    """An abstraction over screen controllers creation.

    By having this abstraction we're solving the problem with circular dependencies between screen
    controllers.
    """

    @abstractmethod
    def get_bundle(self) -> Bundle:
        """..."""

    @abstractmethod
    def get_player_profile(self) -> PlayerProfile:
        """..."""

    @abstractmethod
    def get_main_menu(self) -> ScreenController:
        """Create a new instance of Main Menu screen controller"""

    @abstractmethod
    def get_playfield_screen(self, level_num: int,
                             skip_how_to_play: bool = False) -> ScreenController:
        """Create a new instance of Playfield screen controller"""

    @abstractmethod
    def get_choose_level_screen(self) -> ScreenController:
        """Create a new instance of Choose Level screen controller"""

    @abstractmethod
    def get_game_paused_screen(self, level_num: int) -> ScreenController:
        """Create a new instance of Game Paused screen controller"""

    @abstractmethod
    def get_level_completed_screen(self, level_score: LevelScore) -> ScreenController:
        """Create a new instance of Level Completed screen controller"""

    @abstractmethod
    def get_how_to_play_screen(self) -> ScreenController:
        """Create a new instance of How To Play screen controller"""

    @abstractmethod
    def get_victory_screen(self) -> ScreenController:
        """Create a new instance of Victory Completed screen controller"""

    @abstractmethod
    def get_exit_screen(self, exit_callback: Callable[[], None]) -> ScreenController:
        """Create a new instance of Exit screen controller"""



"""Module exposing graphic related classes and routines."""
from dataclasses import dataclass
from enum import unique, Enum
from functools import total_ordering
from typing import List, Optional, Tuple, Generator

TILEMAP_WIDTH = 256
TILEMAP_HEIGHT = 256
TILE_SIZE = 8

IMAGE_BANK_WIDTH = 256
IMAGE_BANK_HEIGHT = 256

SCREEN_WIDTH = 256
SCREEN_HEIGHT = 256


@unique
class Direction(Enum):
    """Enumeration representing direction in 2D space.

    In addition to int value identifying direction (direction_index) it also stores the
    movement vector (dx, dy)
    """
    UP = 0, (0, -1)
    DOWN = 1, (0, 1)
    LEFT = 2, (-1, 0)
    RIGHT = 3, (1, 0)

    def __init__(self, direction_index: int, delta: Tuple[int, int]) -> None:
        self.direction_index = direction_index
        self.dx = delta[0]
        self.dy = delta[1]

    @property
    def horizontal(self) -> bool:
        """Value indicating whether the direction is horizontal or not."""
        return self in (Direction.LEFT, Direction.RIGHT)

    @property
    def vertical(self) -> bool:
        """Value indicating whether the direction is vertical or not."""
        return self in (Direction.UP, Direction.DOWN)

    @classmethod
    def num_directions(cls) -> int:
        """Number of all defined directions."""
        return len(cls.__members__)

    @property
    def opposite(self) -> "Direction":
        """The opposite direction to given direction."""
        if self == Direction.UP:
            return Direction.DOWN
        if self == Direction.DOWN:
            return Direction.UP
        if self == Direction.LEFT:
            return Direction.RIGHT
        if self == Direction.RIGHT:
            return Direction.LEFT

        raise Exception(f"Direction {str(self)} is not supported")


@dataclass(frozen=True)
class Point:
    """A point representing a location in (x, y) screen space."""
    x: int
    y: int

    @classmethod
    def from_list(cls, coords: List[int]) -> "Point":
        """Create a Point from the list of coordinates."""
        return cls(x=coords[0], y=coords[1])

    @property
    def as_list(self) -> List[int]:
        """Point represented as a list containing [x, y]."""
        return [self.x, self.y]

    def offset(self, offset: "Point") -> "Point":
        """Create a new Point which is the result of moving this Point by (x, y)."""
        return Point(self.x + offset.x, self.y + offset.y)

    def move(self, direction: Direction) -> "Point":
        """Create a new Point which is the result of moving this Point in given direction."""
        return Point(self.x + direction.dx, self.y + direction.dy)


@total_ordering
@dataclass(frozen=True)
class Size:
    """Size describes width and height dimensions in pixels."""
    width: int = 0
    height: int = 0

    def enlarge(self, dx: int, dy: Optional[int] = None) -> "Size":
        """Create a new Size enlarged by (dx, dy)"""
        return Size(self.width + dx, self.height + (dy if dy else dx))

    @property
    def max_dimension(self) -> int:
        """Maximum dimension (which is either width or height)"""
        return max(self.width, self.height)

    def can_fit(self, size: "Size") -> bool:
        """Test if given size will fit in this size. Size "fits" in another size only if both width
        and height are less or equal to the size the test is performed against to.
        """
        return self.width >= size.width and self.height >= size.height

    def __lt__(self, other: Tuple[int, ...]) -> bool:
        return (self.width, self.height) < (other[0], other[1])


def max_size(size1: Size, size2: Size) -> Size:
    """Create size with maximum width and height of two given sizes."""
    return Size(max(size1.width, size2.width), max(size1.height, size2.height))


def min_size(size1: Size, size2: Size) -> Size:
    """Create size with minimum width and height of two given sizes."""
    return Size(min(size1.width, size2.width), min(size1.height, size2.height))


SCREEN_RECT = Size(SCREEN_WIDTH, SCREEN_HEIGHT)


@dataclass(frozen=True)
class Rect:
    """A rectangle represented by position and size.

    Position and size can be accessed directly. Use left, right, top, bottom the get the
    coordinates of rectangle 4 edges.
    """
    position: Point
    size: Size

    @classmethod
    def from_coords(cls, x: int, y: int, w: int, h: int) -> "Rect":
        """Create a new Rect with given position (x, y) and size (w, h)"""
        return cls(position=Point(x, y), size=Size(w, h))

    @classmethod
    def from_list(cls, coords: List[int]) -> "Rect":
        """Create a new Rect from the list containing coordinates and size."""
        return cls(position=Point(coords[0], coords[1]), size=Size(coords[2], coords[3]))

    @classmethod
    def from_size(cls, size: Size) -> "Rect":
        """Create a new Rect with given size and positioned at (0, 0)."""
        return cls(position=Point(0, 0), size=size)

    @property
    def as_list(self) -> List[int]:
        """Rect represented as a list containing [x, y, w, h]."""
        return [self.position.x, self.position.y, self.size.width, self.size.height]

    @property
    def x(self) -> int:
        """The x coordinate of the rect."""
        return self.position.x

    @property
    def y(self) -> int:
        """The y coordinate of the rect."""
        return self.position.y

    @property
    def w(self) -> int:
        """The width of the rect."""
        return self.size.width

    @property
    def h(self) -> int:
        """The height of the rect."""
        return self.size.height

    @property
    def left(self) -> int:
        """The position of left edge of the rect."""
        return self.x

    @property
    def right(self) -> int:
        """The position of right edge of the rect."""
        return self.x + self.w - 1

    @property
    def top(self) -> int:
        """The position of top edge of the rect."""
        return self.y

    @property
    def bottom(self) -> int:
        """The position of bottom edge of the rect."""
        return self.y + self.h - 1

    def offset(self, delta: Point) -> "Rect":
        """Create a new Rect which is the result of moving this Rect by (dx, dy)."""
        return Rect(self.position.offset(delta), self.size)

    def enlarge(self, w: int, h: int) -> "Rect":
        """Create a new Rect enlarged with given size (w, h)."""
        return Rect(position=self.position, size=self.size.enlarge(w, h))

    def inside_points(self) -> Generator[Point, None, None]:
        """Generator for iterating over all valid positions inside the rectangle (from top-left to
        bottom-right)."""
        for y in range(self.y, self.y + self.h):
            for x in range(self.x, self.x + self.w):
                yield Point(x, y)


def hcenter(width: int, target_x: int, target_width: int = SCREEN_WIDTH) -> int:
    """Center horizontally 'size' with specified width in a target section (described by x and
    width)."""
    return target_x + (target_width - width) // 2


def vcenter(height: int, target_y: int, target_height: int = SCREEN_HEIGHT) -> int:
    """Center vertically 'size' with specified height in a target section (described by y and
    height)."""
    return hcenter(height, target_y, target_height)


def center_in_rect(size: Size, target_rect: Rect = Rect.from_size(SCREEN_RECT)) -> Rect:
    """Return rectangle with given size centered in target rectangle."""
    x = hcenter(size.width, target_rect.x, target_rect.w)
    y = vcenter(size.height, target_rect.y, target_rect.h)
    return Rect(Point(x, y), size)


@dataclass(frozen=True)
class Layer:
    """Layer is an abstract surface on which elements can be drawn.

    Layers are used when drawing elements that should be put on each other in order to achieve
    pseudo 3d effect.

    Attributes:
        layer_index - index of the layer (used to calculate layer offset)
        opaque - is layer opaque *OR* transparent
        global_offset - offset of elements drawn on the layer
    """
    layer_index: int
    opaque: bool = False
    global_offset: Point = Point(0, 0)

    @property
    def offset(self) -> Point:
        """Position offset for all graphical objects drawn on this layer."""
        return self.global_offset.offset(Point(-self.layer_index, -self.layer_index))

    @property
    def transparency_color(self) -> int:
        """Transparency color for the layer."""
        return -1 if self.opaque else 0



"""Module exposing tile-related types."""
from dataclasses import dataclass
from enum import Enum, unique, auto


@unique
class TileType(Enum):
    """Enum representing type of a tile in terms of game logic."""
    VOID = auto()
    WALL = auto()
    START = auto()
    FLOOR = auto()
    INITIAL_CRATE_POSITION = auto()
    CRATE_INITIALLY_PLACED = auto()
    CARGO_BAY = auto()

    @property
    def is_start(self) -> bool:
        """Is tile a starting position for the player."""
        return self == TileType.START

    @property
    def is_crate_initially_placed(self) -> bool:
        """Is tile a cargo bay with already placed crate.

        Some crates are already placed in cargo bays at level start.
        """
        return self == TileType.CRATE_INITIALLY_PLACED

    @property
    def is_cargo_bay(self) -> bool:
        """Is tile a cargo bay.

        Cargo bays are the destinations to which crates should be moved to.
        """
        return self == TileType.CARGO_BAY

    @property
    def is_crate_spawn_point(self) -> bool:
        """Is tile a crate spawn point.

        Crate spawn point is a tile at which crate is placed during level initialization.
        """
        return self in (TileType.INITIAL_CRATE_POSITION, TileType.CRATE_INITIALLY_PLACED)

    @property
    def is_walkable(self) -> bool:
        """Is tile a walkable tile.

        A walkable tile means that:
        - player can move to it
        - crates can be pushed to it
        """
        return self in (TileType.START, TileType.FLOOR, TileType.INITIAL_CRATE_POSITION,
                        TileType.CRATE_INITIALLY_PLACED, TileType.CARGO_BAY)


INDEX_TO_TILE = tuple(list(TileType))


@dataclass(frozen=True)
class Tileset:
    """Tileset groups the tiles and provides the mappings for tile types.

    Attributes:
        tileset_index - start index of all tiles from the tileset
    """
    tileset_index: int

    @property
    def first_tile_index(self) -> int:
        """The index of the first tile of the tileset."""
        return self.tileset_index * len(TileType)

    def tile_of(self, tile_index: int) -> TileType:
        """The type of tile with given tile index (which is specific to the tileset)

        :param tile_index: tile index to find tile type for
        :return: type of tile for given tile index in tileset
        """
        num_tiles = len(TileType)
        index_in_range = self.first_tile_index <= tile_index < self.first_tile_index + num_tiles
        return INDEX_TO_TILE[tile_index % num_tiles] if index_in_range else TileType.VOID



"""Bansoko - Space-themed Sokoban clone created in Python using Pyxel.

Usage:
    bansoko [-h] [--version] [--bundle <name>]

Options:
    -h, --help       Show this screen.
    --version        Show version.
    --bundle <name>  Specify resources bundle name [default: main]
"""
import logging
import os
from dataclasses import dataclass
from pathlib import Path

import pyxel
from docopt import docopt

from bansoko import GAME_FRAME_RATE, __version__, GAME_FRAME_TIME_IN_MS
from bansoko.game import GameError
from bansoko.game.bundle import load_bundle, Bundle
from bansoko.game.context import GameContext
from bansoko.game.profile import create_or_load_profile, GAME_PROFILE_LOCATION, \
    GAME_PROFILE_FILENAME, GAME_LOG_FILENAME
from bansoko.game.screens.error import show_error_message
from bansoko.graphics import SCREEN_WIDTH, SCREEN_HEIGHT
from bansoko.gui.navigator import ScreenNavigator

GAME_TITLE = "Bansoko"


@dataclass(frozen=True)
class FileNames:
    """Container for resource and metadata file names."""
    resource_file: str
    metadata_file: str
    profile_file_path: Path
    log_file: str


def generate_filenames(base_name: str) -> FileNames:
    """Generate resource and metadata file names basing on bundle name.

    :param base_name: name of bundle resource and metadata file names are based on
    :return: instance of FileNames
    """
    base_path = Path(os.path.dirname(os.path.realpath(__file__)))
    gamedata_path = base_path.joinpath("gamedata")
    resource_file = gamedata_path.joinpath(base_name + ".pyxres").resolve()
    metadata_file = gamedata_path.joinpath(base_name + ".meta").resolve()
    profile_dir = Path.home().joinpath(GAME_PROFILE_LOCATION)
    os.makedirs(profile_dir, exist_ok=True)
    profile_file_path = profile_dir.joinpath(GAME_PROFILE_FILENAME)
    log_file = profile_dir.joinpath(GAME_LOG_FILENAME)
    return FileNames(str(resource_file), str(metadata_file), profile_file_path, str(log_file))


def configure_logger(log_filename: str) -> None:
    """Sets up a logger for Bansoko."""
    logging.basicConfig(filename=log_filename, filemode="w",
                        format="%(levelname)s%(message)s",
                        level=logging.INFO)
    logging.addLevelName(logging.ERROR, "** ERROR: ")
    logging.addLevelName(logging.WARN, "WARN: ")
    logging.addLevelName(logging.INFO, "")
    logging.info("Starting Bansoko %s", __version__)


def load_game_resources(filenames: FileNames) -> Bundle:
    """Load Pyxel's resource file containing bundle."""
    logging.info("Loading Pyxel resources file '%s'", filenames.resource_file)
    if not os.path.isfile(filenames.resource_file):
        # This is the only way we can pre-check whether pyxel.load() will fail or not
        # In current version of Pyxel it's not possible to react to error or capture the error
        # reason
        raise GameError(f"Unable to find Pyxel resource file '{filenames.resource_file}'")
    pyxel.load(filenames.resource_file)

    logging.info("Loading resources metadata file '%s'", filenames.metadata_file)
    if not os.path.isfile(filenames.metadata_file):
        raise GameError(f"Unable to find resources metadata file '{filenames.metadata_file}'")

    return load_bundle(filenames.metadata_file)


def main() -> None:
    """Main entry point."""
    arguments = docopt(__doc__, version=__version__)
    bundle_name = arguments["--bundle"]
    filenames = generate_filenames(bundle_name)
    configure_logger(filenames.log_file)
    logging.info("Initializing Pyxel window")
    pyxel.init(width=SCREEN_WIDTH, height=SCREEN_HEIGHT, caption=GAME_TITLE, fps=GAME_FRAME_RATE,
               quit_key=pyxel.KEY_F12)
    try:
        bundle = load_game_resources(filenames)
        logging.info("Bundle name: %s", bundle_name)
        logging.info("Bundle SHA1: %s", bundle.sha1.decode())
        player_profile = create_or_load_profile(bundle, filenames.profile_file_path)
        game_context = GameContext(bundle, player_profile)
        navigator = ScreenNavigator(game_context.get_main_menu(), pyxel.quit, GAME_FRAME_TIME_IN_MS)
        logging.info("Game started.")
        pyxel.run(navigator.update, navigator.draw)
    except GameError as error:
        logging.exception(error)
        show_error_message(error.message)


if __name__ == "__main__":
    main()



"""Module for game globals."""


class GameError(Exception):
    """Base class for exceptions in Bansoko."""

    def __init__(self, message: str):
        super().__init__()
        self.message = message



"""Module defining the main game screen."""
from dataclasses import dataclass
from typing import Optional

import pyxel

from bansoko.game.level import InputAction, Level
from bansoko.game.screens.screen_factory import ScreenFactory
from bansoko.graphics import Point, Direction
from bansoko.graphics.animation import AnimationPlayer, Animation
from bansoko.graphics.sprite import Sprite
from bansoko.gui.input import VirtualButton
from bansoko.gui.navigator import ScreenController, BaseScreenController

COCKPIT_LEVEL_NUM_POS = Point(40, 8)
COCKPIT_LEVEL_TIME_POS = Point(174, 10)
COCKPIT_LEVEL_STEPS_POS = Point(145, 227)
COCKPIT_LEVEL_PUSHES_POS = Point(145, 238)
COCKPIT_JOYSTICK_POS = Point(18, 209)
COCKPIT_REWIND_BUTTON_POS = Point(51, 231)
COCKPIT_RECEIPT_POS = Point(202, 202)
COCKPIT_REWIND_ICON_POS = Point(21, 35)
LEFT_HAND_NEUTRAL_POS = Point(-9, 205)
LEFT_HAND_LEFT_POS = Point(-13, 205)
LEFT_HAND_RIGHT_POS = Point(-5, 205)
LEFT_HAND_UP_POS = Point(-7, 203)
LEFT_HAND_DOWN_POS = Point(-11, 208)
LEFT_HAND_PRESSING_BUTTON_POS = Point(19, 234)


@dataclass(frozen=True)
class CockpitElements:
    """Collection of dynamic elements of cockpit."""
    digits_level: Sprite
    digits_steps: Sprite
    digits_pushes: Sprite
    digits_time: Sprite
    rewind_icon: Sprite
    joystick_neutral: Sprite
    joystick_move: Sprite
    rewind_button: Sprite
    left_hand: Sprite
    left_hand_pressing_button: Sprite


class PlayfieldScreen(BaseScreenController):
    """Main game screen controller.

    Screen controller allowing player to "play" the level. It evaluates end-game conditions
    and switches to Level Completed screen when those are met.
    It is also possible to pause the game by pressing either 'Escape' or 'Start'
    (on a gamepad). That switches to Game Paused screen.
    """

    def __init__(self, screen_factory: ScreenFactory, level_num: int,
                 show_how_to_play: bool = False):
        bundle = screen_factory.get_bundle()
        profile = screen_factory.get_player_profile()
        super().__init__(screen=bundle.get_screen("playfield"))
        self.screen_factory = screen_factory
        self.level = Level(bundle.get_level_template(level_num))
        self.ui_elements = CockpitElements(
            digits_level=bundle.get_sprite("digits_yellow"),
            digits_steps=bundle.get_sprite("digits_yellow"),
            digits_pushes=bundle.get_sprite("digits_blue"),
            digits_time=bundle.get_sprite("digits_fat_red"),
            rewind_icon=bundle.get_sprite("rewind_icon"),
            joystick_neutral=bundle.get_sprite("joystick_neutral"),
            joystick_move=bundle.get_sprite("joystick_move"),
            rewind_button=bundle.get_sprite("rewind_button"),
            left_hand=bundle.get_sprite("left_hand"),
            left_hand_pressing_button=bundle.get_sprite("left_hand_pressing_button"))
        self.printing_animation = Animation(bundle.get_sprite("printing_receipt"), 120)
        self.level_completed_player: Optional[AnimationPlayer] = None
        self.how_to_play_shown = not show_how_to_play
        profile.last_played_level = level_num

    def update(self, dt_in_ms: float) -> ScreenController:
        if not self.how_to_play_shown:
            self.how_to_play_shown = True
            return self.screen_factory.get_how_to_play_screen()

        super().update(dt_in_ms)

        if self.level_completed_player:
            return self._update_level_completed_player(dt_in_ms)

        if self.input.is_button_pressed(VirtualButton.START):
            return self.screen_factory.get_game_paused_screen(self.level.level_num)

        if self.level.is_completed:
            return self._start_level_completed_player()

        self.level.process_input(self._get_input_action())
        self.level.update(dt_in_ms)

        return self

    def draw(self, draw_as_secondary: bool = False) -> None:
        pyxel.cls(0)
        if not draw_as_secondary:
            self.level.draw()

        super().draw(draw_as_secondary)
        self._draw_dynamic_cockpit()
        self._draw_level_statistics()
        if self.level_completed_player:
            self.level_completed_player.draw(COCKPIT_RECEIPT_POS)

    def _draw_dynamic_cockpit(self) -> None:
        if not self.level.last_input_action:
            self.ui_elements.joystick_neutral.draw(COCKPIT_JOYSTICK_POS)
            self.ui_elements.rewind_button.draw(COCKPIT_REWIND_BUTTON_POS, frame=0)
            self.ui_elements.left_hand.draw(LEFT_HAND_NEUTRAL_POS)
            return

        input_action = self.level.last_input_action
        if input_action == InputAction.UNDO:
            self.ui_elements.joystick_neutral.draw(COCKPIT_JOYSTICK_POS)
            self.ui_elements.rewind_icon.draw(COCKPIT_REWIND_ICON_POS)
            self.ui_elements.rewind_button.draw(COCKPIT_REWIND_BUTTON_POS, frame=1)
            self.ui_elements.left_hand_pressing_button.draw(LEFT_HAND_PRESSING_BUTTON_POS)
        elif input_action.is_movement:
            self.ui_elements.joystick_move.draw(COCKPIT_JOYSTICK_POS,
                                                direction=input_action.direction)
            if input_action.direction == Direction.LEFT:
                self.ui_elements.left_hand.draw(LEFT_HAND_LEFT_POS)
            elif input_action.direction == Direction.RIGHT:
                self.ui_elements.left_hand.draw(LEFT_HAND_RIGHT_POS)
            elif input_action.direction == Direction.UP:
                self.ui_elements.left_hand.draw(LEFT_HAND_UP_POS)
            elif input_action.direction == Direction.DOWN:
                self.ui_elements.left_hand.draw(LEFT_HAND_DOWN_POS)
            self.ui_elements.rewind_button.draw(COCKPIT_REWIND_BUTTON_POS, frame=0)

    def _draw_level_statistics(self) -> None:
        score = self.level.level_score
        _draw_digits(COCKPIT_LEVEL_NUM_POS, "{:>3d}".format(score.level_num),
                     self.ui_elements.digits_level)
        _draw_digits(COCKPIT_LEVEL_TIME_POS, score.time,
                     self.ui_elements.digits_time, colon_size=4)
        _draw_digits(COCKPIT_LEVEL_STEPS_POS, "{:>4d}".format(score.steps),
                     self.ui_elements.digits_steps)
        _draw_digits(COCKPIT_LEVEL_PUSHES_POS, "{:>4d}".format(score.pushes),
                     self.ui_elements.digits_pushes)

    def _get_input_action(self) -> Optional[InputAction]:
        if self.input.is_button_down(VirtualButton.UP):
            return InputAction.MOVE_UP
        if self.input.is_button_down(VirtualButton.DOWN):
            return InputAction.MOVE_DOWN
        if self.input.is_button_down(VirtualButton.LEFT):
            return InputAction.MOVE_LEFT
        if self.input.is_button_down(VirtualButton.RIGHT):
            return InputAction.MOVE_RIGHT
        if self.input.is_button_down(VirtualButton.ACTION):
            return InputAction.UNDO
        return None

    def _start_level_completed_player(self) -> ScreenController:
        self.level_completed_player = AnimationPlayer(self.printing_animation)
        return self

    def _update_level_completed_player(self, dt_in_ms: float) -> ScreenController:
        if self.level_completed_player:
            if self.level_completed_player.stopped:
                return self.screen_factory.get_level_completed_screen(self.level.level_score)
            self.level_completed_player.update(dt_in_ms)
        return self


def _draw_digits(position: Point, text: str, sprite: Sprite, space: int = 1,
                 colon_size: Optional[int] = None) -> None:
    char_pos = position
    for char in text:
        if char.isdigit():
            sprite.draw(position=char_pos, frame=int(char))
        char_size = colon_size if (char == ":" and colon_size) else sprite.width
        char_pos = char_pos.offset(Point(char_size + space, 0))



"""Module exposing a Bundle, which is a repository of sprites, screens and level templates."""
from dataclasses import dataclass
from json import JSONDecodeError, load
from typing import Dict, Tuple, Any

from jsonschema import validate, ValidationError

from bansoko.game import GameError
from bansoko.game.level_template import LevelTemplate, LevelSpritePacks
from bansoko.game.metadata_schema import METADATA_JSON_SCHEMA
from bansoko.graphics import Rect, Point
from bansoko.graphics.sprite import Sprite, SpritePack
from bansoko.graphics.tilemap import Tilemap
from bansoko.gui.screen import Screen, ScreenElement

SHA1_SIZE_IN_BYTES = 40


@dataclass(frozen=True)
class Bundle:
    """Bundle is a central repository of game resources (such as: sprites, screens
    and level templates)."""

    sha1: bytearray
    sprites: Dict[str, Sprite]
    sprite_packs: Dict[str, SpritePack]
    screens: Dict[str, Screen]
    level_templates: Tuple[LevelTemplate, ...]

    def get_sprite(self, sprite_name: str) -> Sprite:
        """ Return sprite with given sprite name.

        :param sprite_name: name of sprite to be retrieved
        :return: instance of Sprite with given name
        """
        return self.sprites[sprite_name]

    def get_sprite_pack(self, sprite_pack_name: str) -> SpritePack:
        """Return sprite pack with given name.

        :param sprite_pack_name: name of sprite pack to be retrieved
        :return: instance of SpritePack with given name
        """
        return self.sprite_packs[sprite_pack_name]

    def get_screen(self, screen_name: str) -> Screen:
        """Return screen with given screen name.

        :param screen_name: name of screen to be retrieved
        :return: instance of Screen with given name
        """
        return self.screens[screen_name]

    def get_level_template(self, template_id: int) -> LevelTemplate:
        """Return level template with given template id.

        :param template_id: id of template to be retrieved
        :return: instance of LevelTemplate with given id
        """
        return self.level_templates[template_id]

    @property
    def num_levels(self) -> int:
        """Total number of levels in bundle.

        :return: total number of levels bundle contains
        """
        return len(self.level_templates)

    @property
    def last_level(self) -> int:
        """Index of the last level in bundle.

        :return: index of the last level bundle contains
        """
        return self.num_levels - 1


def load_bundle(metadata_filename: str) -> Bundle:
    """Load game resources into bundle using metadata file.

    :param metadata_filename: name of the metadata file
    :return: bundle with game resources
    """
    try:
        with open(metadata_filename) as metadata_file:
            metadata = load(metadata_file)
            validate(metadata, METADATA_JSON_SCHEMA)
            sprites = create_sprites(metadata["sprites"])
            sprite_packs = create_sprite_packs(metadata["sprite_packs"], sprites)
            screens = create_screens(metadata["screens"], sprites)
            sha1 = bytearray(metadata["levels"]["sha1"], "utf-8").zfill(
                SHA1_SIZE_IN_BYTES)[-SHA1_SIZE_IN_BYTES:]
            level_templates = create_level_templates(
                metadata["levels"]["level_templates"], sprite_packs)
            return Bundle(sha1, sprites, sprite_packs, screens, level_templates)
    except JSONDecodeError as decode_error:
        raise GameError("Incorrect format of resource metadata file") from decode_error
    except ValidationError as validation_error:
        raise GameError("Incorrect format of resource metadata file") from validation_error


def create_sprites(json_data: Any) -> Dict[str, Sprite]:
    """Create sprites from metadata.

    :param json_data: input JSON containing sprites metadata
    :return: collection of sprites
    """
    return {
        name: Sprite(
            image_bank=data["image_bank"],
            uv_rect=Rect.from_list(data["uv_rect"]),
            directional=data["directional"],
            transparency_color=data["transparency_color"],
            num_layers=data["num_layers"],
            num_frames=data["num_frames"])
        for name, data in json_data.items()
    }


def create_sprite_packs(json_data: Any, sprites: Dict[str, Sprite]) -> Dict[str, SpritePack]:
    """Create sprite packs from metadata.

    :param json_data: input JSON containing sprite packs metadata
    :param sprites: collection of available sprites
    :return: collection of sprite packs
    """
    return {
        name: SpritePack(
            sprites=tuple([sprites[sprite_name] for sprite_name in data]))
        for (name, data) in json_data.items()
    }


def create_screens(json_data: Any, sprites: Dict[str, Sprite]) -> Dict[str, Screen]:
    """Create screens from metadata.

    :param json_data: input JSON containing screens metadata
    :param sprites: collection of available sprites
    :return: collection of screens
    """
    return {
        name: _screen_from_json(data, sprites)
        for (name, data) in json_data.items()
    }


def _screen_from_json(json_data: Any, sprites: Dict[str, Sprite]) -> Screen:
    background_data = json_data.get("background")
    background_color = None
    background_tilemap = None
    if background_data:
        background_color = background_data.get("background_color")
        tilemap_data = background_data.get("background_tilemap")
        if tilemap_data:
            background_tilemap = Tilemap(
                tilemap_id=tilemap_data["tilemap_id"],
                rect_uv=Rect.from_list(tilemap_data["tilemap_uv"])
            )

    screen_elements = []
    if json_data.get("elements"):
        screen_elements = [
            ScreenElement(
                position=Point.from_list(data["position"]),
                sprite=sprites.get(data.get("sprite_ref")),
                text=data.get("text"))
            for data in json_data["elements"]
        ]
    menu_position = None
    menu_scrollbar_rect = None
    if json_data.get("menu"):
        menu_data = json_data["menu"]
        menu_position = \
            Point.from_list(menu_data["position"]) if menu_data.get("position") else None
        menu_scrollbar_rect = \
            Rect.from_list(menu_data["scrollbar_rect"]) if menu_data.get("scrollbar_rect") else None

    return Screen(
        background_color=background_color, background_tilemap=background_tilemap,
        elements=tuple(screen_elements), menu_position=menu_position,
        menu_scrollbar_rect=menu_scrollbar_rect)


def create_level_templates(json_data: Any, sprite_packs: Dict[str, SpritePack]) \
        -> Tuple[LevelTemplate, ...]:
    """Create level templates from metadata.

    :param json_data: input JSON containing level template metadata
    :param sprite_packs: collection of available sprite packs
    :return: collection of level templates
    """
    return tuple([
        LevelTemplate.from_level_num(
            level_num=level_num,
            tileset_index=data["tileset"],
            draw_offset=Point.from_list(data["draw_offset"]),
            sprite_packs=LevelSpritePacks(
                robot_sprite_pack=sprite_packs[data["robot_sprite_pack_ref"]],
                crate_sprite_pack=sprite_packs[data["crate_sprite_pack_ref"]]))
        for level_num, data in enumerate(json_data)])



"""Module defining game screen containing information on how to play."""

from bansoko.game.screens.screen_factory import ScreenFactory
from bansoko.gui.menu import MenuController, TextMenuItem, Menu, MenuLayout


class HowToPlayController(MenuController):
    """Screen controller for displaying how to play instructions.

    Player can only go back to playfield screen from here.
    """

    def __init__(self, screen_factory: ScreenFactory):
        screen = screen_factory.get_bundle().get_screen("how_to_play")
        menu = Menu.with_defaults(tuple([
            TextMenuItem("OK", lambda: None)
        ]), MenuLayout(position=screen.menu_position))
        super().__init__(menu=menu, allow_going_back=True, screen=screen, semi_transparent=True)



"""Module exposing animation related classes."""
from dataclasses import dataclass
from typing import Optional

from bansoko.graphics import Point, Direction, Layer
from bansoko.graphics.sprite import Sprite


@dataclass(frozen=True)
class Animation:
    """Animation exposes information needed for playing a sequence of frames stored in a sprite.

    Attributes:
        sprite - sprite containing animation frame
        frame_length - the length of the single frame (expressed in ms)
        looped - should animation be played from the beginning after reaching the last frame
    """
    sprite: Sprite
    frame_length: float
    looped: bool = False

    @property
    def animation_length(self) -> float:
        """Total animation length (expressed in ms)."""
        return self.num_frames * self.frame_length

    @property
    def num_frames(self) -> int:
        """Total number of frames in the animation."""
        return self.sprite.num_frames

    @property
    def last_frame(self) -> int:
        """Index of the last frame of the animation."""
        return self.num_frames - 1

    def frame_at_time(self, animation_time: float) -> int:
        """Return index of the animation frame corresponding to given animation time.

        :param animation_time: animation time to find frame at (expressed in ms)
        :return: the index of animation frame corresponding to given animation time
        """
        frame = round(animation_time / self.frame_length)
        return frame % self.num_frames if self.looped else min(self.last_frame, frame)

    def stopped_at_time(self, animation_time: float) -> bool:
        """Check if the animation will be stopped at given animation time.

        :param animation_time: animation time to be checked
        :return: True if animation will be stopped (not played) at given time; False otherwise
        """
        return not self.looped and animation_time > self.animation_length


@dataclass
class AnimationPlayer:
    """Player for animations.

    Call to AnimationPlayer's update() is required in each frame.

    Attributes:
        animation - animation that is played by the animation player
        current_frame - the current frame of the animation (according to animation_time)
        animation_time - elapsed animation time (updated in each call to update())
    """
    animation: Optional[Animation] = None
    current_frame: int = 0
    animation_time: float = 0.0

    def play(self, animation: Animation) -> None:
        """Set up the animation player to play given animation.

        :param animation: animation to be played
        """
        self.animation = animation
        self.current_frame = 0
        self.animation_time = 0.0

    def update(self, dt_in_ms: float) -> None:
        """Update animation player with time that elapsed over a single game frame.

        It updates total animation time and calculates the current animation frame.

        :param dt_in_ms: delta time since last update (in ms)
        """
        if not self.animation:
            return

        self.animation_time += dt_in_ms
        self.current_frame = self.animation.frame_at_time(self.animation_time)

    def draw(self, position: Point, layer: Optional[Layer] = None,
             direction: Direction = Direction.UP) -> None:
        """Draw current animation frame at given position using specified layer and direction.

        Current frame is calculated based on elapsed animation time.

        :param position: position of animation frame to be drawn at
        :param layer: layer of animation frame to be drawn at
        :param direction: direction-specific variant of animation frame to be drawn
        """
        if not self.animation:
            return

        self.animation.sprite.draw(position, layer=layer, direction=direction,
                                   frame=self.current_frame)

    @property
    def stopped(self) -> bool:
        """Is animation player stopped.

        Animation player is stopped when it has no animation *OR* animation is not looped and was
        played already.
        """
        return not self.animation or self.animation.stopped_at_time(self.animation_time)



"""Module for game screens management."""
import abc
from typing import Optional, Callable, List

from bansoko.gui.input import InputSystem
from bansoko.gui.screen import Screen


class ScreenController(abc.ABC):
    """Base class for all game screen controllers that suppose to be managed by ScreenNavigator.

    Screen controllers are updated and drawn once per frame. Screen transitions are triggered by
    values returned in update method.
    """

    def __init__(self, semi_transparent: Optional[bool] = False) -> None:
        self.semi_transparent = semi_transparent

    @abc.abstractmethod
    def activate(self) -> None:
        """Called each time screen controller is put on top of screen stack by ScreenNavigator."""

    @abc.abstractmethod
    def update(self, dt_in_ms: float) -> Optional["ScreenController"]:
        """Update screen controller state. Control screen transitions by returning value.

        Called once per frame (only if screen is on top of screen stack)

        :param dt_in_ms: delta time since last update (in ms)
        :return: self (or any instance of type(self)) - no screen transition *OR*
        instance of ScreenController class - switch to new screen controller *OR*
        None - switch to previous screen controller (perform pop on screen stack)
        """

    @abc.abstractmethod
    def draw(self, draw_as_secondary: bool = False) -> None:
        """Draw screen.

        Called once per frame (only if screen controller is on top of screen stack)

        :param draw_as_secondary: is this screen drawn as a secondary (background) screen
        """


class BaseScreenController(ScreenController):
    """Base class for all game screen controllers that suppose to be managed by ScreenNavigator.

    Screen controllers are updated and drawn once per frame. Screen transitions are triggered by
    values returned in update method.
    """

    def __init__(self, semi_transparent: Optional[bool] = False,
                 screen: Optional[Screen] = None) -> None:
        super().__init__(semi_transparent)
        self.screen = screen
        self.input = InputSystem()

    def activate(self) -> None:
        self.input.reset()

    def update(self, dt_in_ms: float) -> Optional["ScreenController"]:
        self.input.update()
        return self

    def draw(self, draw_as_secondary: bool = False) -> None:
        if self.screen is not None:
            self.screen.draw()


class ScreenNavigator:
    """ScreenNavigator manages game screen controllers.

    Game screen controllers are organized in a stack. Screen controller from the top is called an
    active screen controller. Active screen controller receives update and draw calls once per
    frame.
    Active screen controller can trigger:
        - switch to new screen controller (which will be put on top of the stack and activated)
        - end its life and switch to previous screen controller (controller from top will be popped
          from the stack and then new controller from top will be activated)
    Switching between screen controllers is controlled by update() callback from ScreenController
    class.
    """

    def __init__(self, start_controller: ScreenController, exit_callback: Callable[[], None],
                 frame_time: float):
        self.controllers_stack = [start_controller]
        self.exit_callback = exit_callback
        self.frame_time = frame_time
        self.skip_next_draw = False
        start_controller.activate()

    def update(self) -> None:
        """Update screen controller from top of controllers stack. Manage screen transitions."""
        if self.controllers_stack:
            new_screen = self.controllers_stack[-1].update(self.frame_time)
            if not isinstance(new_screen, type(self.controllers_stack[-1])):
                self._switch_to_screen(new_screen)
        else:
            self.exit_callback()
            self.skip_next_draw = True

    def draw(self) -> None:
        """Draw screen controller from top of controllers stack."""
        if not self.skip_next_draw:
            screens_to_be_drawn: List[ScreenController] = []
            for screen in reversed(self.controllers_stack):
                screens_to_be_drawn.insert(0, screen)
                if not screen.semi_transparent:
                    break

            for (i, screen) in enumerate(screens_to_be_drawn):
                is_top_screen = (i == len(screens_to_be_drawn) - 1)
                screen.draw(draw_as_secondary=not is_top_screen)
        self.skip_next_draw = False

    def _switch_to_screen(self, new_screen: Optional[ScreenController]) -> None:
        if new_screen is None:
            self.controllers_stack.pop()
        else:
            self._unwind_screen_stack(new_screen)
            self.controllers_stack.append(new_screen)
        if self.controllers_stack:
            self.controllers_stack[-1].activate()
        self.skip_next_draw = True

    def _unwind_screen_stack(self, screen_to_unwind_to: ScreenController) -> None:
        for i, screen in enumerate(self.controllers_stack):
            if isinstance(screen, type(screen_to_unwind_to)):
                del self.controllers_stack[i:]
                break



"""Module defining main menu screen controller."""
from typing import Optional, Callable

from bansoko import __version__
from bansoko.game.screens.screen_factory import ScreenFactory
from bansoko.graphics import Point
from bansoko.graphics.text import draw_text, TextStyle
from bansoko.gui.menu import MenuController, TextMenuItem, Menu, MenuLayout
from bansoko.gui.navigator import ScreenController


class MainMenuController(MenuController):
    """Main Menu for the game (entry point for the game).

    Screen controller allowing player to start a new game, choose a level from a list of
    all unlocked levels or exit the game.
    """

    def __init__(self, screen_factory: ScreenFactory):
        self.exiting = False
        self._get_exit_screen: Callable[[], ScreenController] = \
            lambda: screen_factory.get_exit_screen(self._exit_callback)
        level_to_play = screen_factory.get_player_profile().first_not_completed_level
        screen = screen_factory.get_bundle().get_screen("main_menu")
        menu = Menu.with_defaults((
            TextMenuItem("START GAME", lambda: screen_factory.get_playfield_screen(level_to_play)),
            TextMenuItem("CHOOSE LEVEL", screen_factory.get_choose_level_screen),
            TextMenuItem("EXIT", self._get_exit_screen)
        ), MenuLayout(position=screen.menu_position))
        super().__init__(menu=menu, allow_going_back=True, screen=screen)

    def activate(self) -> None:
        super().activate()
        self.select_and_scroll_to_item(0)

    def draw(self, draw_as_secondary: bool = False) -> None:
        super().draw(draw_as_secondary)
        draw_text(Point(79, 240), "(c) 2020 KRZYSZTOF FURTAK", TextStyle(color=7, shadow_color=1))
        draw_text(Point(11, 240), f"v{__version__}", TextStyle(color=1))

    def update(self, dt_in_ms: float) -> Optional[ScreenController]:
        if self.exiting:
            return None
        controller_to_go_to = super().update(dt_in_ms)
        return controller_to_go_to if controller_to_go_to else self._get_exit_screen()

    def _exit_callback(self) -> None:
        self.exiting = True



"""Module exposing GUI elements like screen, image and text."""
from dataclasses import dataclass
from typing import Optional, Tuple

import pyxel

from bansoko.graphics import Point, Layer, Rect
from bansoko.graphics.sprite import Sprite
from bansoko.graphics.text import draw_text
from bansoko.graphics.tilemap import Tilemap


@dataclass(frozen=True)
class ScreenElement:
    """ScreenElement is a drawable part of Screen.

    It's described by sprite *OR* text and by position to be drawn at.
    """
    position: Point
    sprite: Optional[Sprite] = None
    text: Optional[str] = None

    def draw(self) -> None:
        """Draw screen element at its defined position."""
        if self.sprite:
            self.sprite.draw(self.position)
        elif self.text:
            draw_text(self.position, self.text)


@dataclass(frozen=True)
class Screen:
    """Screen is a composition of elements that can be drawn on a game screen."""
    background_color: Optional[int] = None
    background_tilemap: Optional[Tilemap] = None
    elements: Tuple[ScreenElement, ...] = ()
    menu_position: Optional[Point] = None
    menu_scrollbar_rect: Optional[Rect] = None

    def draw(self) -> None:
        """Draw screen with all its elements.

        If background color is specified then screen is cleared with that color first.
        """
        if self.background_color is not None:
            pyxel.cls(self.background_color)
        if self.background_tilemap:
            self.background_tilemap.draw(Layer(0))
        if self.elements:
            for element in self.elements:
                element.draw()



"""Module for game menus management."""
from abc import ABC, abstractmethod
from dataclasses import dataclass
from functools import reduce
from itertools import islice
from typing import Callable, Optional, Iterable, Tuple

from bansoko.graphics import Size, Point, max_size, center_in_rect
from bansoko.graphics.text import draw_text, text_size, TextStyle
from bansoko.gui.input import VirtualButton
from bansoko.gui.navigator import ScreenController, BaseScreenController
from bansoko.gui.screen import Screen


class MenuItem(ABC):
    """An abstract, base class for all menu items controlled by MenuScreen."""

    def __init__(self, controller_to_switch_to: Callable[[], Optional[ScreenController]]):
        self.controller_to_switch_to = controller_to_switch_to

    @property
    @abstractmethod
    def disabled(self) -> bool:
        """Is menu item disabled.

        It is not possible to perform action on disabled menu items.
        """

    @property
    @abstractmethod
    def size(self) -> Size:
        """Size (in pixels) of the menu item.

        This property is used during layout
        """

    @abstractmethod
    def draw(self, position: Point, selected: bool = False) -> None:
        """Draw menu item at given position.

        :param position: position to draw menu item at
        :param selected: should the item be drawn as selected
        """

    def perform_action(self) -> Optional[ScreenController]:
        """Perform action tied up to the menu item.

        Navigation between game screens is controlled by return value.

        :return: instance of Screen class - switch to new screen controller *OR*
        None - switch to previous screen controller (exit menu screen)
        """
        return self.controller_to_switch_to()


MENU_TEXT_NORMAL = TextStyle(color=7, shadow_color=1)
MENU_TEXT_SELECTED = TextStyle(color=10, shadow_color=1)
MENU_TEXT_DISABLED = TextStyle(color=12, shadow_color=1)


class TextMenuItem(MenuItem):
    """Text-based menu item.

    It contains only label, which changes color when item is selected.
    """

    def __init__(self, text: str,
                 controller_to_switch_to: Callable[[], Optional[ScreenController]]):
        super().__init__(controller_to_switch_to)
        self.text = text

    @property
    def disabled(self) -> bool:
        return False

    @property
    def size(self) -> Size:
        return text_size(self._get_item_text(selected=True), MENU_TEXT_NORMAL).enlarge(2)

    def draw(self, position: Point, selected: bool = False) -> None:
        if self.disabled:
            style = MENU_TEXT_DISABLED
        elif selected:
            style = MENU_TEXT_SELECTED
        else:
            style = MENU_TEXT_NORMAL
        draw_text(position, self._get_item_text(selected), style)

    def _get_item_text(self, selected: bool = False) -> str:
        return ("* " if selected else "  ") + self.text


@dataclass(frozen=True)
class MenuLayout:
    """Menu layout configuration.

    Attributes:
        columns - number of visible columns in a single row of the menu
        rows - number of visible rows in the menu (if None it will be defaulted to total rows)
        position - screen-space position of the menu (if None menu will be centered on screen)
                   if one of the coordinates (x or y) is equal to -1 then menu will be centered
                   on screen in appropriate axis (horizontal or vertical)
        item_space - space between items in the menu
    """

    columns: int = 1
    rows: Optional[int] = None
    position: Optional[Point] = None
    item_space: Size = Size(0, 0)


@dataclass(frozen=True)
class Menu:
    """Menu is a list of options, user can navigate and choose from.

    Attributes:
        items - collection of menu items
        item_size - standardised size of menu item
        item_space - space between items in the menu
        columns - number of visible columns in a single row of the menu
        rows - number of visible rows in the menu
        position - screen-space position of the menu
    """

    items: Tuple[MenuItem, ...]
    item_size: Size
    item_space: Size
    columns: int
    rows: int
    position: Point

    @classmethod
    def with_defaults(cls, items: Tuple[MenuItem, ...],
                      layout: MenuLayout = MenuLayout()) -> "Menu":
        """Construct menu based on defaults.

        :param items: collection of menu items
        :param layout: layout information of the menu
        :return: newly created instance of Menu
        """
        item_size = reduce(max_size, [item.size for item in items])
        total_rows = -(-len(items) // layout.columns)
        calculated_rows = min(layout.rows, total_rows) if layout.rows else total_rows
        calculated_size = Size(
            layout.columns * item_size.width + (layout.columns - 1) * layout.item_space.width,
            calculated_rows * item_size.height + (calculated_rows - 1) * layout.item_space.height)
        centered_position = center_in_rect(calculated_size).position

        if layout.position:
            x = layout.position.x if layout.position.x >= 0 else centered_position.x
            y = layout.position.y if layout.position.y >= 0 else centered_position.y
            menu_position = Point(x, y)
        else:
            menu_position = centered_position

        return Menu(
            items=items,
            item_size=item_size,
            item_space=layout.item_space,
            columns=layout.columns,
            rows=calculated_rows,
            position=menu_position)

    @property
    def total_rows(self) -> int:
        """Total number of rows in the menu."""
        return -(-len(self.items) // self.columns)

    @property
    def items_on_page(self) -> int:
        """Number of items visible on single menu page."""
        return self.columns * self.rows


class MenuController(BaseScreenController):
    """MenuController is a screen controller containing configurable menu.

    Attributes:
        items - collection of all menu items that are going to be
        item_size - calculated maximum size of all menu items
        config - configuration the menu is initialized with
        position - screen-space position of the menu
        selected_item - index of currently selected menu item
        top_row - the very firs, top row that is visible when menu is drawn (it changes during
                  scrolling)
    """

    def __init__(self, menu: Menu, allow_going_back: bool = False,
                 screen: Optional[Screen] = None, semi_transparent: bool = False) -> None:
        super().__init__(semi_transparent=semi_transparent, screen=screen)
        self.menu = menu
        self.allow_going_back = allow_going_back
        self.top_row = 0
        self.selected_item = 0

    @property
    def scrollbar_size(self) -> float:
        """Size of the scroll bar relative to total number of menu rows.

        :return: Value from 0.0 to 1.0 describing the relation of the number of visible rows to
                 the number of all rows in the menu
        """
        return self.menu.rows / self.menu.total_rows

    @property
    def scrollbar_position(self) -> float:
        """Position of the scroll bar relative to total number of menu rows.

        :return: Value from 0.0 to 1.0 describing the relation of first visible row (top row) to
                 the number of all rows in the menu
        """
        return self.top_row / self.menu.total_rows

    @property
    def visible_items(self) -> Iterable[MenuItem]:
        """Collection of all visible menu items with respect to scroll bar position."""
        top_left_item = self.top_row * self.menu.columns
        bottom_left_item = min(
            (self.top_row + self.menu.rows) * self.menu.columns, len(self.menu.items))
        return islice(self.menu.items, top_left_item, bottom_left_item)

    def select_and_scroll_to_item(self, item: int) -> None:
        """Select menu item with given number and scroll menu to be sure that it's visible.

        :param item: item to be selected
        """
        if item < 0:
            self.selected_item = 0
        elif item >= len(self.menu.items):
            self.selected_item = len(self.menu.items) - 1
        else:
            self.selected_item = item

        self.scroll_to_item(self.selected_item)

    def scroll_to_item(self, item: int) -> None:
        """Scroll menu to specified item."""
        self.top_row = min(item // self.menu.columns, self.menu.total_rows - self.menu.rows)

    def update(self, dt_in_ms: float) -> Optional[ScreenController]:
        super().update(dt_in_ms)

        if self.input.is_button_pressed(VirtualButton.BACK) and self.allow_going_back:
            return None

        if not self.menu.items:
            return self

        selected_item_disabled = self.menu.items[self.selected_item].disabled
        if self.input.is_button_pressed(VirtualButton.SELECT) and not selected_item_disabled:
            return self.menu.items[self.selected_item].perform_action()

        if self.input.is_button_pressed(VirtualButton.UP):
            self._move_selection_up()
        elif self.input.is_button_pressed(VirtualButton.DOWN):
            self._move_selection_down()
        elif self.input.is_button_pressed(VirtualButton.LEFT):
            self._move_selection_left()
        elif self.input.is_button_pressed(VirtualButton.RIGHT):
            self._move_selection_right()
        elif self.input.is_button_pressed(VirtualButton.HOME):
            self._select_first()
        elif self.input.is_button_pressed(VirtualButton.END):
            self._select_last()
        elif self.input.is_button_pressed(VirtualButton.PAGE_UP):
            self._move_selection_page_up()
        elif self.input.is_button_pressed(VirtualButton.PAGE_DOWN):
            self._move_selection_page_down()

        return self

    def draw(self, draw_as_secondary: bool = False) -> None:
        super().draw(draw_as_secondary)

        for i, item in enumerate(self.visible_items):
            item_space = self.menu.item_space
            calculated_item_size = self.menu.item_size.enlarge(item_space.width, item_space.height)
            position = self.menu.position.offset(Point(
                (i % self.menu.columns) * calculated_item_size.width,
                (i // self.menu.columns) * calculated_item_size.height))
            item.draw(position, (self.top_row * self.menu.columns) + i == self.selected_item)

    def _move_selection_up(self) -> None:
        selected_row = self.selected_item // self.menu.columns
        move_up_possible = (selected_row > 0)
        if move_up_possible:
            self.selected_item = self.selected_item - self.menu.columns
            self.top_row = min(self.top_row, selected_row - 1)

    def _move_selection_down(self) -> None:
        selected_row = self.selected_item // self.menu.columns
        move_down_possible = (selected_row < (len(self.menu.items) - 1) // self.menu.columns)
        if move_down_possible:
            self.selected_item = min(
                self.selected_item + self.menu.columns, len(self.menu.items) - 1)
            if selected_row + 1 >= self.top_row + self.menu.rows:
                self.top_row += 1

    def _move_selection_left(self) -> None:
        selected_column = self.selected_item % self.menu.columns
        move_left_possible = (selected_column > 0)
        if move_left_possible:
            self.selected_item -= 1

    def _move_selection_right(self) -> None:
        selected_column = self.selected_item % self.menu.columns
        move_right_possible = (selected_column < self.menu.columns - 1) and \
                              (self.selected_item + 1 < len(self.menu.items))
        if move_right_possible:
            self.selected_item += 1

    def _select_first(self) -> None:
        self.select_and_scroll_to_item(0)

    def _select_last(self) -> None:
        self.select_and_scroll_to_item(len(self.menu.items) - 1)

    def _move_selection_page_up(self) -> None:
        self.select_and_scroll_to_item(self.selected_item - self.menu.items_on_page)

    def _move_selection_page_down(self) -> None:
        self.select_and_scroll_to_item(self.selected_item + self.menu.items_on_page)



"""Module exposing Sprite type."""
from dataclasses import dataclass
from typing import Tuple, Optional

import pyxel

from bansoko.graphics import Rect, Point, Direction, Layer


@dataclass(frozen=True)
class Sprite:
    """Sprite is a part of an Image that can be drawn on screen at given position.

    The exact fragment of the image is defined by image_bank and uv_rect (which is a rectangle in
    image space)
    Additionally, sprite can be:
    - multilayer - it contains variants for all layers,
    - directional - it contains variants for all 4 directions,
    - transparent - it is drawn with transparency color (-1 means opaque)
    - multiframe - it contains multiple frames (for example for animation purposes)

    Please note that those features can be combined, so we can have sprite which is multilayer,
    directional and multiframe at the same time.
    """
    image_bank: int
    uv_rect: Rect
    directional: bool = False
    transparency_color: int = -1
    num_layers: int = 1
    num_frames: int = 1

    def draw(self, position: Point, layer: Optional[Layer] = None,
             direction: Direction = Direction.UP, frame: int = 0) -> None:
        """Draw the sprite at given position using specified layer, direction and frame.

        If sprite is single-layered then it will be drawn _only on main_ layer.

        :param position: position of sprite to be drawn at
        :param layer: layer of sprite to be drawn at
        :param direction: direction-specific variant of sprite to be drawn
        :param frame: frame of sprite to be drawn
        """
        if layer and layer.layer_index >= self.num_layers:
            return

        clamped_frame = min(frame, self.num_frames - 1)
        frame_offset_v = clamped_frame * self.uv_rect.h // self.num_frames
        top_layer_offset = self.num_layers - 1

        u = self.uv_rect.x + top_layer_offset
        v = self.uv_rect.y + frame_offset_v + top_layer_offset

        if self.directional:
            u += self.uv_rect.w // Direction.num_directions() * direction.direction_index

        if layer:
            u -= layer.layer_index
            v -= layer.layer_index

        offset = layer.offset if layer else Point(0, 0)

        pyxel.blt(position.x + offset.x, position.y + offset.y, self.image_bank,
                  u, v, self.width, self.height, self.transparency_color)

    @property
    def width(self) -> int:
        """The width of sprite in pixels."""
        width = self.uv_rect.w

        if self.directional:
            width //= Direction.num_directions()

        width -= self.num_layers - 1

        return width

    @property
    def height(self) -> int:
        """The height of sprite in pixels."""
        return self.uv_rect.h // self.num_frames - (self.num_layers - 1)


@dataclass(frozen=True)
class SpritePack:
    """Sprite Pack is a collection of sprites, grouped together for organizational purposes."""
    sprites: Tuple[Sprite, ...]



"""Module defining actions that can be executed during game."""
import abc
from typing import Optional

from bansoko import GAME_FRAME_TIME_IN_MS
from bansoko.game.game_object import GameStats, GameObject, Robot, RobotState, Crate
from bansoko.graphics import Direction, Point, TILE_SIZE


class GameAction(abc.ABC):
    """GameAction is a base, abstract class for all game actions.

    Attributes:
        time_to_complete - time after which the action will be completed (in ms)
        elapsed_time - time the action already took (in ms)
        chain_action - secondary, child action that runs along with the action
        backward - is the action running in backward mode (True when we're undo-ing action)
    """

    def __init__(self, time_to_complete: float,
                 chain_action: Optional["GameAction"] = None) -> None:
        self.time_to_complete = time_to_complete
        self.chain_action = chain_action
        self.elapsed_time = 0.0
        self.backward = False

    def update(self, dt_in_ms: float, stats: GameStats) -> Optional["GameAction"]:
        """Update action with given delta time.

        Called once per frame.

        :param dt_in_ms: delta time since last update (in ms)
        :param stats: statistics the action can contribute to
        :return: None if action has complete *OR* self (or any instance of type(self)) otherwise
        """
        if self.chain_action:
            self.chain_action.update(dt_in_ms, stats)
        self.elapsed_time += dt_in_ms
        if self.elapsed_time >= self.time_to_complete:
            self._on_complete(stats)
            return None
        return self

    def reset(self, backward: bool = False) -> None:
        """Reset the action, so it can be executed again.

        :param backward: backward mode the action will be run with (True if we're undo-ing action)
        """
        if self.chain_action:
            self.chain_action.reset(backward=backward)
        self.elapsed_time = 0.0
        self.backward = backward

    def _on_complete(self, stats: GameStats) -> None:
        pass


class MoveAction(GameAction):
    """MoveAction is an abstract class responsible for handling game object movement in tilemap."""

    def __init__(self, game_object: GameObject, direction: Direction,
                 time_to_complete: float, chain_action: Optional[GameAction] = None) -> None:
        super().__init__(time_to_complete, chain_action)
        self.game_object = game_object
        self.direction = direction

    def update(self, dt_in_ms: float, stats: GameStats) -> Optional[GameAction]:
        running_action = super().update(dt_in_ms, stats)
        if running_action:
            move_direction = self.direction.opposite if self.backward else self.direction
            delta = self.elapsed_time / self.time_to_complete * TILE_SIZE
            self.game_object.position.offset = Point(
                int(delta * move_direction.dx), int(delta * move_direction.dy))

        return running_action

    def _on_complete(self, stats: GameStats) -> None:
        move_direction = self.direction.opposite if self.backward else self.direction
        self.game_object.position.move(move_direction)


class TurnRobot(GameAction):
    """TurnRobot is an action for turing the robot in given direction.

    Attributes:
        robot - robot to be turned
        direction - direction to turn the robot in
    """

    def __init__(self, robot: Robot, direction: Direction) -> None:
        super().__init__(time_to_complete=0.0)
        self.robot = robot
        self.direction = direction

    def update(self, dt_in_ms: float, stats: GameStats) -> Optional[GameAction]:
        running_action = super().update(dt_in_ms, stats)
        self.robot.face_direction = self.direction
        return running_action

    def reset(self, backward: bool = False) -> None:
        super().reset(backward=backward)
        self.robot.face_direction = self.direction


TIME_TO_COMPLETE_ROBOT_MOVE = TILE_SIZE * GAME_FRAME_TIME_IN_MS


class MoveRobot(MoveAction):
    """MoveRobot is a move action that encapsulates the movement of robot."""

    def __init__(self, robot: Robot, direction: Direction,
                 move_state: RobotState = RobotState.MOVING):
        super().__init__(robot, direction, TIME_TO_COMPLETE_ROBOT_MOVE, TurnRobot(robot, direction))
        self.robot = robot
        self.move_state = move_state
        self.robot.robot_state = move_state

    def reset(self, backward: bool = False) -> None:
        super().reset(backward)
        self.robot.robot_state = self.move_state
        # TODO: What about animation when moving backward?

    def _on_complete(self, stats: GameStats) -> None:
        super()._on_complete(stats)
        stats.steps += 1 if not self.backward else -1


TIME_TO_COMPLETE_CRATE_PUSH = TILE_SIZE * GAME_FRAME_TIME_IN_MS


class PushCrate(MoveAction):
    """PushCrate is a move action that encapsulates the movement of robot and the push of crate."""

    def __init__(self, robot: Robot, crate: Crate, direction: Direction) -> None:
        super().__init__(crate, direction, TIME_TO_COMPLETE_CRATE_PUSH,
                         chain_action=MoveRobot(robot, direction, RobotState.PUSHING))
        self.crate = crate

    def _on_complete(self, stats: GameStats) -> None:
        super()._on_complete(stats)
        stats.pushes += 1 if not self.backward else -1



"""Module defining screen controller for choosing a level to be played."""

import pyxel

from bansoko import LEVEL_THUMBNAIL_IMAGE_BANK, LEVEL_WIDTH, LEVEL_HEIGHT
from bansoko.game.screens.screen_factory import ScreenFactory
from bansoko.graphics import Point, Size
from bansoko.graphics.text import draw_text, TextStyle
from bansoko.gui.menu import MenuController, MenuItem, Menu, MenuLayout

LEVEL_SELECTED_STYLE = TextStyle(color=10)
LEVEL_LOCKED_STYLE = TextStyle(color=1)
LEVEL_UNLOCKED_STYLE = TextStyle(color=5)
LEVEL_COMPLETED_STYLE = TextStyle(color=3)

LEVEL_ITEM_SIZE = Size(38, 45)
LEVEL_ITEM_SPACE = Size(6, 6)
LEVEL_ITEM_TITLE_POS = Point(3, 3)
LEVEL_LOCKED_ICON_POS = Point(3, 10)
LEVEL_COMPLETED_ICON_POS = Point(3, 38)
LEVEL_THUMBNAIL_POS = Point(3, 10)
LEVEL_SCORE_POS = Point(14, 237)


class LevelMenuItem(MenuItem):
    """LevelMenuItem represents a menu item used for selecting level on ChooseLevel screen."""

    def __init__(self, level_num: int, screen_factory: ScreenFactory):
        super().__init__(lambda: screen_factory.get_playfield_screen(level_num))
        self.level_num = level_num
        self.player_profile = screen_factory.get_player_profile()
        self.check_icon = screen_factory.get_bundle().get_sprite("check_icon")
        self.locked_icon = screen_factory.get_bundle().get_sprite("locked_icon")

    @property
    def disabled(self) -> bool:
        return self.level_num > self.player_profile.last_unlocked_level

    @property
    def size(self) -> Size:
        return LEVEL_ITEM_SIZE

    @property
    def text_style(self) -> TextStyle:
        """Text style of the menu item.

        It depends on the status of level completion the menu item is referring to.
        """
        if self.player_profile.is_level_completed(self.level_num):
            return LEVEL_COMPLETED_STYLE
        if self.player_profile.is_level_unlocked(self.level_num):
            return LEVEL_UNLOCKED_STYLE

        return LEVEL_LOCKED_STYLE

    @property
    def level_unlocked(self) -> bool:
        """Does this menu item represent already unlocked level."""
        return self.player_profile.is_level_unlocked(self.level_num)

    @property
    def level_completed(self) -> bool:
        """Does this menu item represent completed level."""
        return self.player_profile.is_level_completed(self.level_num)

    def draw(self, position: Point, selected: bool = False) -> None:
        draw_text(position.offset(LEVEL_ITEM_TITLE_POS), f"LEVEL {self.level_num}", self.text_style)

        if not self.level_unlocked:
            self.locked_icon.draw(position.offset(LEVEL_LOCKED_ICON_POS))
        else:
            self._draw_level_thumbnail(position.offset(LEVEL_THUMBNAIL_POS))

        if self.level_completed:
            self.check_icon.draw(position.offset(LEVEL_COMPLETED_ICON_POS))

        if selected:
            self._draw_level_score(LEVEL_SCORE_POS)

        self._draw_frame(position, selected)

    def _draw_level_thumbnail(self, position: Point) -> None:
        pyxel.blt(position.x, position.y, LEVEL_THUMBNAIL_IMAGE_BANK,
                  LEVEL_WIDTH * (self.level_num % 8),
                  LEVEL_HEIGHT * (self.level_num // 8),
                  LEVEL_WIDTH, LEVEL_HEIGHT, colkey=0)

    def _draw_level_score(self, position: Point) -> None:
        level_score = self.player_profile.levels_scores[self.level_num]

        if self.level_completed:
            text = f"TIME: {level_score.time}  " \
                   f"PUSHES: {level_score.pushes: >4}  " \
                   f"STEPS: {level_score.steps: >4}"
        else:
            text = "LEVEL NOT COMPLETED" if self.level_unlocked else "LEVEL LOCKED"

        draw_text(position, text, LEVEL_SELECTED_STYLE)

    def _draw_frame(self, position: Point, selected: bool) -> None:
        style = LEVEL_SELECTED_STYLE if selected else self.text_style
        pyxel.rectb(position.x, position.y, LEVEL_ITEM_SIZE.width, LEVEL_ITEM_SIZE.height,
                    style.color)


class ChooseLevelController(MenuController):
    """Screen controller allowing player to choose a level to play.

    Player can choose a level from all unlocked levels (Level is "unlocked"
    when its predecessor level is completed)
    From this screen it is also possible to navigate back to Main Menu.
    """

    def __init__(self, screen_factory: ScreenFactory):
        bundle = screen_factory.get_bundle()
        screen = bundle.get_screen("choose_level")
        menu = Menu.with_defaults(tuple([
            LevelMenuItem(level_num, screen_factory) for level_num in range(bundle.num_levels)
        ]), MenuLayout(columns=5, rows=4, position=screen.menu_position,
                       item_space=LEVEL_ITEM_SPACE))
        super().__init__(menu=menu, allow_going_back=True, screen=screen)
        self.select_and_scroll_to_item(screen_factory.get_player_profile().last_played_level)

    def draw(self, draw_as_secondary: bool = False) -> None:
        super().draw(draw_as_secondary)
        self._draw_scroll_bar()

    def _draw_scroll_bar(self) -> None:
        if not self.screen or not self.screen.menu_scrollbar_rect:
            return

        scrollbar_rect = self.screen.menu_scrollbar_rect
        scrollbar_size_in_pixels = round(super().scrollbar_size * scrollbar_rect.h)
        scrollbar_position_in_pixels = round(super().scrollbar_position * scrollbar_rect.h)
        pyxel.rect(scrollbar_rect.x, scrollbar_rect.y + scrollbar_position_in_pixels,
                   scrollbar_rect.w, scrollbar_size_in_pixels, LEVEL_SELECTED_STYLE.color)



"""Module defining game screen which is displayed when level is completed."""
from typing import Tuple

from bansoko.game.profile import LevelScore
from bansoko.game.screens.screen_factory import ScreenFactory
from bansoko.graphics import Point
from bansoko.graphics.text import draw_text
from bansoko.gui.menu import MenuController, TextMenuItem, MenuItem, Menu, MenuLayout

LEVEL_TIME_POS = Point(104, 75)
LEVEL_PUSHES_POS = Point(104, 84)
LEVEL_STEPS_POS = Point(104, 93)


class LevelCompletedController(MenuController):
    """Screen controller allowing player to choose what to do after completion of the level.

    Player can navigate to the next level, replay the current level
    (to get better score) or get back to Main Menu.
    """

    def __init__(self, screen_factory: ScreenFactory, level_score: LevelScore):
        current_level_num = level_score.level_num
        last_level_completed = current_level_num == screen_factory.get_bundle().last_level

        next_level = TextMenuItem(
            "PLAY NEXT LEVEL", lambda: screen_factory.get_playfield_screen(
                screen_factory.get_player_profile().next_level_to_play(current_level_num)))
        finish_game = TextMenuItem(
            "FINISH GAME", screen_factory.get_victory_screen)
        restart_level = TextMenuItem(
            "RESTART LEVEL", lambda: screen_factory.get_playfield_screen(
                current_level_num, skip_how_to_play=True))
        main_menu = TextMenuItem(
            "BACK TO MAIN MENU", screen_factory.get_main_menu)

        more_levels_to_play_menu: Tuple[MenuItem, ...] = (next_level, restart_level, main_menu)
        last_level_finished_menu: Tuple[MenuItem, ...] = (finish_game, restart_level)
        screen = screen_factory.get_bundle().get_screen("level_completed")
        menu = Menu.with_defaults(
            last_level_finished_menu if last_level_completed else more_levels_to_play_menu,
            MenuLayout(position=screen.menu_position))

        super().__init__(menu=menu, screen=screen, semi_transparent=True)
        player_profile = screen_factory.get_player_profile()
        self.level_score = level_score
        self.prev_level_score = player_profile.complete_level(level_score)

    def draw(self, draw_as_secondary: bool = False) -> None:
        super().draw(draw_as_secondary)
        self._draw_level_statistics()

    def _draw_level_statistics(self) -> None:
        new_record = "(NEW RECORD)"
        time_beaten = self.level_score.time_in_ms < self.prev_level_score.time_in_ms
        pushes_beaten = self.level_score.pushes < self.prev_level_score.pushes
        steps_beaten = self.level_score.steps < self.prev_level_score.steps
        first_completion = not self.prev_level_score.completed

        draw_text(LEVEL_TIME_POS, "#0{:>7s} #8{:s}".format(
            self.level_score.time, new_record if time_beaten or first_completion else ""))
        draw_text(LEVEL_PUSHES_POS, "#0{:>7d} #8{:s}".format(
            self.level_score.pushes, new_record if pushes_beaten or first_completion else ""))
        draw_text(LEVEL_STEPS_POS, "#0{:>7d} #8{:s}".format(
            self.level_score.steps, new_record if steps_beaten or first_completion else ""))



"""Module defining screen controller which is displayed before exiting the game."""
from typing import Callable

from bansoko.game.screens.screen_factory import ScreenFactory
from bansoko.graphics import Size
from bansoko.gui.menu import MenuController, TextMenuItem, Menu, MenuLayout


class ExitController(MenuController):
    """Screen controller for displaying exit confirmation dialog."""

    def __init__(self, screen_factory: ScreenFactory, exit_callback: Callable[[], None]):
        screen = screen_factory.get_bundle().get_screen("exit")
        menu = Menu.with_defaults(tuple([
            TextMenuItem("YES", self._exit),
            TextMenuItem("NO", lambda: None)
        ]), MenuLayout(columns=2, position=screen.menu_position, item_space=Size(8, 0)))
        super().__init__(menu=menu, allow_going_back=True, screen=screen, semi_transparent=True)
        self.exit_callback = exit_callback

    def _exit(self) -> None:
        self.exit_callback()



"""Module exposing all game objects."""
import abc
from dataclasses import dataclass
from enum import IntEnum, unique
from typing import Dict

from bansoko.graphics import Point, Direction, Layer
from bansoko.graphics.animation import AnimationPlayer, Animation
from bansoko.graphics.sprite import Sprite
from bansoko.graphics.tilemap import TilePosition


@dataclass
class GameStats:
    """Statistics of currently played game.

    Attributes:
        game_time - total time spent on playing the level (in ms)
        pushes - number of moves that player made
        steps - number of steps that player made
    """
    game_time = 0.0
    pushes: int = 0
    steps: int = 0


@dataclass
class ObjectPosition:
    """Position of game object in the level.

    Attributes:
        tile_position - game object position in tilemap space
        offset - position offset relative to tile_position (expressed in pixels)
    """
    tile_position: TilePosition
    offset: Point = Point(0, 0)

    def move(self, direction: Direction) -> None:
        """Move object position in given direction by one tile.

        Additionally reset offset.

        :param direction: direction of the movement
        """
        self.tile_position = self.tile_position.move(direction)
        self.offset = Point(0, 0)

    def to_point(self) -> Point:
        """Convert tile position to a point in screen space (taking into account the offset)."""
        return self.tile_position.to_point().offset(self.offset)


class GameObject(abc.ABC):
    """This is an abstract, base class for all game objects.

    Attributes:
        position - game object position
    """
    position: ObjectPosition

    def __init__(self, tile_position: TilePosition) -> None:
        self.position = ObjectPosition(tile_position)

    @property
    def tile_position(self) -> TilePosition:
        """The position in tilemap this game object is located at."""
        return self.position.tile_position

    def update(self, dt_in_ms: float) -> None:
        """Update game object.

        Called once per frame.

        :param dt_in_ms: delta time since last update (in ms)
        """

    @abc.abstractmethod
    def draw(self, layer: Layer) -> None:
        """Draw game object on given layer.

        Called once per frame.

        :param layer: layer the game object is drawn on
        """


@unique
class RobotState(IntEnum):
    """Enum defining all states robot can be in.

    When no action is performed robot is in STANDING state. From there it can either transition to
    PUSHING state (when there is a crate that can be moved in robot moving direction) or MOVING
    state otherwise.
    """
    STANDING = 0
    MOVING = 1
    PUSHING = 2


class Robot(GameObject):
    """Robot is the main game object, controlled by the player.

    It can move through the level and push crates.

    Attributes:
        face_direction - direction the robot is facing to
        robot_animations - animation pack for all robot states
        animation_player - used for playing robot animations (animation player updated by Robot)
        _robot_state - internal robot's state (used for picking the right animation)
    """
    _robot_state: RobotState

    def __init__(self, tile_position: TilePosition, face_direction: Direction,
                 robot_animations: Dict[RobotState, Animation]):
        super().__init__(tile_position)
        self.face_direction = face_direction
        self.robot_animations = robot_animations
        self.animation_player = AnimationPlayer()
        self.robot_state = RobotState.STANDING

    @property
    def robot_state(self) -> RobotState:
        """The current state of the robot.

        Change of state results in a change of robot animation.
        """
        return self._robot_state

    @robot_state.setter
    def robot_state(self, robot_state: RobotState) -> None:
        self.animation_player.play(self.robot_animations[robot_state])
        self._robot_state = robot_state

    def update(self, dt_in_ms: float) -> None:
        self.animation_player.update(dt_in_ms)

    def draw(self, layer: Layer) -> None:
        self.animation_player.draw(self.position.to_point(), layer, self.face_direction)


@unique
class CrateState(IntEnum):
    """Enum defining all states crate can be in.

    Crates can be either PLACED (when they are located in cargo bays) or MISPLACED otherwise.
    """
    MISPLACED = 0
    PLACED = 1


class Crate(GameObject):
    """Crate represents a crate game object that can be moved and placed in cargo bays (which is
    the goal of the game)

    Attributes:
        crate_sprites - collection of sprites for all crate states
        state - the state crate is currently in (used for picking the right sprite to draw with and
                importantly to determine whether a crate is in place or not)
    """

    def __init__(self, tile_position: TilePosition, is_placed: bool,
                 crate_sprites: Dict[CrateState, Sprite]) -> None:
        super().__init__(tile_position)
        self.crate_sprites = crate_sprites
        self.state = CrateState.PLACED if is_placed else CrateState.MISPLACED

    @property
    def in_place(self) -> bool:
        """Is the crate placed on a cargo bay tile."""
        return self.state == CrateState.PLACED

    def draw(self, layer: Layer) -> None:
        sprite = self.crate_sprites[self.state]
        sprite.draw(self.position.to_point(), layer)



#!/usr/bin/env python
import os

from setuptools import setup, find_packages

import bansoko

setup(
    name="bansoko",
    version=bansoko.__version__,
    author="Krzysztof Furtak",
    author_email="contact@krzysztoffurtak.dev",
    url="https://github.com/kfurtak1024/bansoko",
    description="Space-themed Sokoban clone created in Python using Pyxel.",
    long_description=open(
        os.path.join(os.path.abspath(os.path.dirname(__file__)), "README.md")
    ).read(),
    long_description_content_type="text/markdown",
    platforms="any",
    keywords="game puzzle sokoban 8-bit retro",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3 :: Only",
        "Topic :: Games/Entertainment :: Puzzle Games"
    ],
    install_requires=["setuptools", "pyxel", "docopt", "jsonschema"],
    packages=find_packages(exclude=["resbuilder", "resbuilder.*"]),
    package_data={
        "bansoko": ["gamedata/main.pyxres", "gamedata/main.meta"]
    },
    entry_points={"console_scripts": ["bansoko = bansoko.__main__:main"]}
)



"""Module exposing PlayerProfile, that can read/write information about game progress."""
import logging
import os
from dataclasses import dataclass
from pathlib import Path
from typing import BinaryIO, List

from bansoko.game import GameError
from bansoko.game.bundle import Bundle, SHA1_SIZE_IN_BYTES

GAME_PROFILE_LOCATION = ".bansoko"
GAME_PROFILE_FILENAME = "profile.data"
GAME_LOG_FILENAME = "bansoko.log"

FILE_HEADER = bytes.fromhex("42 41 4E 53 01")
INITIALLY_UNLOCKED_LEVEL = 2
INT_SIZE_IN_BYTES = 4
LEVEL_SCORE_SIZE_IN_BYTES = 4 * INT_SIZE_IN_BYTES


@dataclass(frozen=True)
class LevelScore:
    """Player's score for given level.

    Attributes:
        level_num - level number
        completed - has level been completed
        pushes - number of moves that player made ("move" happens when player pushes a crate)
        steps - number of steps that player made ("step" happens when player moves by one cell in
                any direction)
        time_in_ms - time spent playing the level (expressed in milliseconds)
    """
    level_num: int
    completed: bool = False
    pushes: int = 0
    steps: int = 0
    time_in_ms: int = 0

    @property
    def time(self) -> str:
        """Human readable level completion time expressed in format H:MM:SS."""

        total_seconds = self.time_in_ms // 1000
        total_minutes = total_seconds // 60
        total_hours = total_minutes // 60
        hours = total_hours % 60
        minutes = total_minutes % 60
        seconds = total_seconds % 60

        if self.time_in_ms >= 10 * 60 * 60 * 1000:
            hours = 9
            seconds = 59
            minutes = 59

        return "{:d}:{:02d}:{:02d}".format(hours, minutes, seconds)

    def merge_with(self, level_score: "LevelScore") -> "LevelScore":
        """Merge this level score with given score.

        In case of differences in field values prefer better score
        (smaller number of steps, pushes, quicker completion time)

        :param level_score: level score to merge with
        :return: newly created level score which is a result of merge
        """
        if self == level_score:
            return level_score
        if self.level_num != level_score.level_num:
            raise GameError("Cannot merge scores from different levels")

        if not self.completed:
            return level_score

        return LevelScore(
            level_num=self.level_num,
            completed=self.completed,
            pushes=min(self.pushes, level_score.pushes),
            steps=min(self.steps, level_score.steps),
            time_in_ms=min(self.time_in_ms, level_score.time_in_ms))


class PlayerProfile:
    """PlayerProfile is a storage for keeping player's game progress.

    Attributes:
        _profile_file_path - path of the player profile file
        _file_offset - position in profile file where level scores start (in bytes)
        _last_unlocked_level - index of last unlocked level
        levels_scores - list of scores for all levels
        last_played_level - last level played by player (not persisted)
    """

    def __init__(self, profile_file_path: Path, file_offset: int, last_unlocked_level: int,
                 levels_scores: List[LevelScore]):
        self._profile_file_path = profile_file_path
        self._file_offset = file_offset
        self._last_unlocked_level = last_unlocked_level
        self.levels_scores = levels_scores
        self.last_played_level = 0

    @property
    def first_not_completed_level(self) -> int:
        """The first not completed level.

        This is the level that is going to be started when player chooses 'START GAME' option in
        main menu.

        :return: The first not completed level if there is one *OR* 0 otherwise
        """
        level = next((level for level in self.levels_scores if not level.completed),
                     self.levels_scores[0])
        return level.level_num

    @property
    def last_unlocked_level(self) -> int:
        """The last unlocked level. The last level from all levels that can be played now."""
        return self._last_unlocked_level

    @property
    def last_level(self) -> int:
        """The last level.."""
        return len(self.levels_scores) - 1

    def is_level_unlocked(self, level_num: int) -> bool:
        """Test if specified level is unlocked (which means that it can be played now)

        :param level_num: level to be tested
        :return: True - if level is unlocked *OR* False - otherwise
        """
        return level_num <= self._last_unlocked_level

    def is_level_completed(self, level_num: int) -> bool:
        """Test if specified level was ever completed.

        :param level_num: level to be tested
        :return: True - if level was completed *OR* False - otherwise
        """
        return self.levels_scores[level_num].completed

    def next_level_to_play(self, level_num: int) -> int:
        """Get the next level to be played after completion of given level."""
        if self._last_unlocked_level == self.last_level:
            return (level_num + 1) % (self.last_level + 1)

        future_levels = self.levels_scores[level_num + 1:self.last_unlocked_level + 1]
        next_level = next((level.level_num for level in future_levels if not level.completed),
                          self.first_not_completed_level)
        return next_level

    def complete_level(self, level_score: LevelScore) -> LevelScore:
        """Save information about level completion to profile file. Additionally, as a reward,
        unlock next level.

        :param level_score: score of level completion
        :return: the previous score for the completed level
        """
        logging.info("Updating player profile file with game progress")

        if not self.is_level_completed(level_score.level_num):
            level_to_be_unlocked = self._last_unlocked_level + 1
            if self._can_unlock_level(level_to_be_unlocked):
                self._last_unlocked_level = level_to_be_unlocked

        prev_level_score = self.levels_scores[level_score.level_num]
        new_level_score = prev_level_score.merge_with(level_score)
        self.levels_scores[level_score.level_num] = new_level_score

        try:
            with open(self._profile_file_path, "r+b") as profile_file:
                profile_file.seek(self._file_offset)
                _write_int(profile_file, self._last_unlocked_level)
                profile_file.seek(new_level_score.level_num * LEVEL_SCORE_SIZE_IN_BYTES, 1)
                _write_int(profile_file, 1 if new_level_score.completed else 0)
                _write_int(profile_file, new_level_score.steps)
                _write_int(profile_file, new_level_score.pushes)
                _write_int(profile_file, new_level_score.time_in_ms)
        except IOError as io_error:
            raise GameError(
                f"Unable to update player profile file '{self._profile_file_path}'. "
                "Progress lost :-(") from io_error

        return prev_level_score

    def _can_unlock_last_level(self) -> bool:
        return next(level.level_num for level in self.levels_scores if
                    not level.completed) == self.last_level

    def _can_unlock_level(self, level_to_be_unlocked: int) -> bool:
        if level_to_be_unlocked > self.last_level:
            return False

        return level_to_be_unlocked != self.last_level or self._can_unlock_last_level()


def create_or_load_profile(bundle: Bundle, profile_file_path: Path) -> PlayerProfile:
    """Create or load (if already exists) a player profile.

    :param bundle: bundle the profile should be initialized with
    :param profile_file_path: path to player profile file
    :return: initialized player profile
    """
    if not os.path.isfile(profile_file_path):
        return _create_profile_file(profile_file_path, bundle)

    return _load_profile_file(profile_file_path, bundle)


def _create_profile_file(profile_file_path: Path, bundle: Bundle) -> PlayerProfile:
    try:
        logging.info("Creating new player profile file '%s'", profile_file_path)
        with open(profile_file_path, "wb") as profile_file:
            profile_file.write(FILE_HEADER)
            return _init_player_profile(profile_file_path, profile_file, bundle)
    except IOError as io_error:
        raise GameError("Unable to create player profile file") from io_error


def _load_profile_file(profile_file_path: Path, bundle: Bundle) -> PlayerProfile:
    try:
        logging.info("Loading existing player profile file '%s'", profile_file_path)
        with open(profile_file_path, "r+b") as profile_file:
            header = profile_file.read(len(FILE_HEADER))
            if header != FILE_HEADER:
                raise GameError("File is not a valid player profile file")

            while True:
                sha1 = profile_file.read(SHA1_SIZE_IN_BYTES)
                if not sha1:
                    break

                section_size = _read_int(profile_file)

                if sha1 == bundle.sha1:
                    return _read_player_profile(profile_file_path, profile_file, bundle)

                profile_file.seek(section_size, 1)

            return _init_player_profile(profile_file_path, profile_file, bundle)
    except IOError as io_error:
        raise GameError("Unable to open player profile file") from io_error
    except EOFError as eof_error:
        raise GameError("Unexpected end of player profile file") from eof_error


def _write_int(file: BinaryIO, value: int) -> None:
    file.write(value.to_bytes(INT_SIZE_IN_BYTES, byteorder="big"))


def _write_zeros(file: BinaryIO, count: int) -> None:
    file.write((0).to_bytes(count, byteorder="big"))


def _init_player_profile(profile_file_path: Path, profile_file: BinaryIO,
                         bundle: Bundle) -> PlayerProfile:
    profile_file.write(bundle.sha1)
    _write_int(profile_file, INT_SIZE_IN_BYTES + bundle.num_levels * LEVEL_SCORE_SIZE_IN_BYTES)

    file_offset = profile_file.tell()
    last_unlocked_level = INITIALLY_UNLOCKED_LEVEL
    levels_scores = [LevelScore(level_num) for level_num in range(bundle.num_levels)]

    _write_int(profile_file, last_unlocked_level)
    _write_zeros(profile_file, bundle.num_levels * LEVEL_SCORE_SIZE_IN_BYTES)

    return PlayerProfile(profile_file_path, file_offset, last_unlocked_level, levels_scores)


def _read_int(file: BinaryIO) -> int:
    int_from_file = file.read(INT_SIZE_IN_BYTES)
    if len(int_from_file) != INT_SIZE_IN_BYTES:
        raise EOFError()
    return int.from_bytes(int_from_file, byteorder="big")


def _read_player_profile(profile_file_path: Path, profile_file: BinaryIO,
                         bundle: Bundle) -> PlayerProfile:
    levels_scores = []
    file_offset = profile_file.tell()
    last_unlocked_level = _read_int(profile_file)
    for level_num in range(bundle.num_levels):
        levels_scores.append(
            LevelScore(
                level_num=level_num,
                completed=_read_int(profile_file) > 0,
                steps=_read_int(profile_file),
                pushes=_read_int(profile_file),
                time_in_ms=_read_int(profile_file)))

    return PlayerProfile(profile_file_path, file_offset, last_unlocked_level,
                         levels_scores)



"""Module defining screen controller which is displayed when game is paused."""
from typing import Tuple

from bansoko.game.screens.screen_factory import ScreenFactory
from bansoko.gui.menu import MenuController, TextMenuItem, Menu, MenuLayout, MenuItem


class GamePausedController(MenuController):
    """Screen controller for displaying pause menu in the game.

    From this screen player can resume the game, restart the current level
    (restoring level to its initial state) display How To Play instructions or return back to
    the Main Menu.
    """

    def __init__(self, screen_factory: ScreenFactory, level_num: int):
        screen = screen_factory.get_bundle().get_screen("game_paused")
        playing_last_level = level_num == screen_factory.get_bundle().last_level
        resume_game = TextMenuItem("RESUME GAME", lambda: None)
        restart_level = TextMenuItem("RESTART LEVEL", lambda: screen_factory.get_playfield_screen(
            level_num, skip_how_to_play=True))
        skip_level = TextMenuItem("SKIP LEVEL", lambda: screen_factory.get_playfield_screen(
            screen_factory.get_player_profile().next_level_to_play(level_num)))
        how_to_play = TextMenuItem("HOW TO PLAY", screen_factory.get_how_to_play_screen)
        main_menu = TextMenuItem("BACK TO MAIN MENU", screen_factory.get_main_menu)

        before_last_level_menu: Tuple[MenuItem, ...] = \
            (resume_game, restart_level, skip_level, how_to_play, main_menu)
        last_level_menu: Tuple[MenuItem, ...] = \
            (resume_game, restart_level, how_to_play, main_menu)

        menu = Menu.with_defaults(
            last_level_menu if playing_last_level else before_last_level_menu,
            MenuLayout(position=screen.menu_position))
        super().__init__(menu=menu, allow_going_back=True, screen=screen, semi_transparent=True)

    def draw(self, draw_as_secondary: bool = False) -> None:
        if not draw_as_secondary:
            super().draw(draw_as_secondary=draw_as_secondary)



"""Module defining game context shared between all game screens."""
from typing import Callable

from bansoko.game.bundle import Bundle
from bansoko.game.profile import PlayerProfile, LevelScore
from bansoko.game.screens.choose_level import ChooseLevelController
from bansoko.game.screens.exit import ExitController
from bansoko.game.screens.game_paused import GamePausedController
from bansoko.game.screens.how_to_play import HowToPlayController
from bansoko.game.screens.level_completed import LevelCompletedController
from bansoko.game.screens.main_menu import MainMenuController
from bansoko.game.screens.playfield import PlayfieldScreen
from bansoko.game.screens.screen_factory import ScreenFactory
from bansoko.game.screens.victory import VictoryController
from bansoko.gui.navigator import ScreenController


class GameContext(ScreenFactory):
    """GameContext is a screen factory that is shared between all game screens."""

    def __init__(self, bundle: Bundle, player_profile: PlayerProfile):
        self.bundle = bundle
        self.player_profile = player_profile

    def get_bundle(self) -> Bundle:
        return self.bundle

    def get_player_profile(self) -> PlayerProfile:
        return self.player_profile

    def get_main_menu(self) -> ScreenController:
        return MainMenuController(self)

    def get_playfield_screen(self, level_num: int,
                             skip_how_to_play: bool = False) -> ScreenController:
        skip_how_to_play = (level_num == 0) and not skip_how_to_play
        return PlayfieldScreen(self, level_num, show_how_to_play=skip_how_to_play)

    def get_choose_level_screen(self) -> ScreenController:
        return ChooseLevelController(self)

    def get_game_paused_screen(self, level_num: int) -> ScreenController:
        return GamePausedController(self, level_num)

    def get_level_completed_screen(self, level_score: LevelScore) -> ScreenController:
        return LevelCompletedController(self, level_score)

    def get_how_to_play_screen(self) -> ScreenController:
        return HowToPlayController(self)

    def get_victory_screen(self) -> ScreenController:
        return VictoryController(self)

    def get_exit_screen(self, exit_callback: Callable[[], None]) -> ScreenController:
        return ExitController(self, exit_callback)



"""Module exposing basic system for input handling."""
from enum import unique, IntFlag
from typing import Dict, List, Set

import pyxel


@unique
class VirtualButton(IntFlag):
    """Virtual button of a virtual game controller."""

    UP = 0x001
    DOWN = 0x002
    LEFT = 0x004
    RIGHT = 0x008
    SELECT = 0x010
    BACK = 0x020
    START = 0x040
    ACTION = 0x080
    HOME = 0x100
    END = 0x200
    PAGE_UP = 0x400
    PAGE_DOWN = 0x800


class InputSystem:
    """InputSystem is a wrapper around Pyxel's input handling.

    It operates on VirtualButton which is an abstraction over physical buttons.
    InputSystem needs to be updated by calling update() method in each frame.
    """

    KEY_HOLD_TIME: int = 10
    KEY_PERIOD_TIME: int = 2
    BUTTONS_MAP: Dict[VirtualButton, List[int]] = {
        VirtualButton.UP: [pyxel.KEY_UP, pyxel.KEY_KP_8, pyxel.GAMEPAD_1_UP],
        VirtualButton.DOWN: [pyxel.KEY_DOWN, pyxel.KEY_KP_2, pyxel.GAMEPAD_1_DOWN],
        VirtualButton.LEFT: [pyxel.KEY_LEFT, pyxel.KEY_KP_4, pyxel.GAMEPAD_1_LEFT],
        VirtualButton.RIGHT: [pyxel.KEY_RIGHT, pyxel.KEY_KP_6, pyxel.GAMEPAD_1_RIGHT],
        VirtualButton.SELECT: [pyxel.KEY_ENTER, pyxel.KEY_KP_ENTER, pyxel.GAMEPAD_1_A],
        VirtualButton.BACK: [pyxel.KEY_ESCAPE, pyxel.GAMEPAD_1_B],
        VirtualButton.START: [pyxel.KEY_ESCAPE, pyxel.GAMEPAD_1_START],
        VirtualButton.ACTION: [pyxel.KEY_Z, pyxel.KEY_BACKSPACE, pyxel.GAMEPAD_1_B],
        VirtualButton.HOME: [pyxel.KEY_HOME],
        VirtualButton.END: [pyxel.KEY_END],
        VirtualButton.PAGE_UP: [pyxel.KEY_PAGE_UP],
        VirtualButton.PAGE_DOWN: [pyxel.KEY_PAGE_DOWN]
    }
    WATCHED_KEYS: Set[int] = set(sum(BUTTONS_MAP.values(), []))

    pressed_keys: Dict[int, int]

    def __init__(self) -> None:
        self.pressed_keys = {}

    def is_button_pressed(self, button: VirtualButton) -> bool:
        """Test if given virtual button is "pressed".

        Button is "pressed" when:
            - in previous update frame it was up and now it's down *OR*
            - it is down for exactly KEY_HOLD_TIME (and was preceded be previous point) *OR*
            - in every KEY_PERIOD_TIME interval (and was preceded be previous point)

        Note that above is valid only if it was not interrupted by a reset call.
        Both KEY_HOLD_TIME and KEY_PERIOD_TIME are expressed in number of occurred update frames.

        :param button: virtual button to be tested
        :return: True - if button was pressed *OR* False - otherwise
        """
        return any(self._is_key_pressed(key) for key in self.BUTTONS_MAP[button])

    def is_button_down(self, button: VirtualButton) -> bool:
        """Test if given virtual button is down at the current update frame.

        :param button: virtual button to be tested
        :return: True - if button is down in current update frame *OR* False - otherwise
        """
        return any(self._is_key_down(key) for key in self.BUTTONS_MAP[button])

    def is_button_up(self, button: VirtualButton) -> bool:
        """Test if given virtual button is up at the current update frame.

        :param button: virtual button to be tested
        :return: True - if button is up in current update frame *OR* False - otherwise
        """
        return not self.is_button_down(button)

    def update(self) -> None:
        """Update statuses of all watched keys.

        This needs to be called each frame.
        """
        for key in self.WATCHED_KEYS:
            if pyxel.btn(key):
                if key not in self.pressed_keys and not pyxel.btnp(key):
                    return
                self._press_key(key)
            else:
                self._depress_key(key)

    def reset(self) -> None:
        """Reset the statuses of all watched keys.

        This should be called when focus switches to another screen.
        """
        self.pressed_keys.clear()

    def _is_key_pressed(self, key: int) -> bool:
        if pyxel.btn(pyxel.KEY_ALT):
            return False
        if key not in self.pressed_keys:
            return False
        hold_time = self.pressed_keys[key]
        if hold_time == 0:
            return True
        if hold_time < self.KEY_HOLD_TIME:
            return False
        if (hold_time - self.KEY_HOLD_TIME) % self.KEY_PERIOD_TIME == 0:
            return True
        return False

    def _is_key_down(self, key: int) -> bool:
        if pyxel.btn(pyxel.KEY_ALT):
            return False
        return key in self.pressed_keys

    def _press_key(self, key: int) -> None:
        self.pressed_keys[key] = self.pressed_keys.get(key, -1) + 1

    def _depress_key(self, key: int) -> None:
        if key in self.pressed_keys:
            del self.pressed_keys[key]



"""Module containing level related classes."""
from enum import Enum
from itertools import chain
from typing import Optional, List, Iterable

from bansoko.game.game_action import GameAction, PushCrate, MoveRobot, TurnRobot
from bansoko.game.game_object import GameObject, Crate, RobotState, CrateState, GameStats
from bansoko.game.level_template import LevelTemplate
from bansoko.game.profile import LevelScore
from bansoko.graphics import Direction
from bansoko.graphics.tilemap import TilePosition


class InputAction(Enum):
    """InputAction represents all possible actions player can perform when playing a level."""
    MOVE_LEFT = Direction.LEFT
    MOVE_RIGHT = Direction.RIGHT
    MOVE_UP = Direction.UP
    MOVE_DOWN = Direction.DOWN
    UNDO = None

    def __init__(self, direction: Direction) -> None:
        self.direction = direction

    @property
    def is_movement(self) -> bool:
        """Is this action a movement action."""
        return self.direction is not None


class Level:
    """Level encapsulates the game logic for playing a level.

    It draws the level with all game objects, it handles the input (used for controlling Robot) and
    it evaluates victory condition.

    Attributes:
        statistics - statistics captures during the play of the level
        template - template the level is crated from
        robot - instance of Robot game object, that player can control
        crates - collection of all Crate game objects for the level
        running_action - currently running game action (updated in update method)
        last_input_action - input action that triggered running_action
        history - list of historical game actions (used for undo)
    """

    def __init__(self, template: LevelTemplate) -> None:
        self.statistics = GameStats()
        self.template = template
        self.robot = template.create_robot()
        self.crates = template.create_crates()
        self.running_action: Optional[GameAction] = None
        self.last_input_action: Optional[InputAction] = None
        self.history: List[GameAction] = []

    @property
    def level_score(self) -> LevelScore:
        """Current player's score in the level."""
        return LevelScore(self.template.level_num, self.is_completed, self.statistics.pushes,
                          self.statistics.steps, int(self.statistics.game_time))

    @property
    def level_num(self) -> int:
        """The number of the level."""
        return self.template.level_num

    @property
    def is_completed(self) -> bool:
        """Test if level objectives are completed.

        Level is completed when all crates are in cargo bays.
        """
        return not next((crate for crate in self.crates if not crate.in_place), None)

    @property
    def game_objects(self) -> Iterable[GameObject]:
        """Collection of all game objects."""
        return chain([self.robot], self.crates)

    def crate_at_pos(self, position: TilePosition) -> Optional[Crate]:
        """Test if there is a crate at given position.

        :param position: position to test the presence of crate at
        :return: True - if there is a crate in given position *OR* False - otherwise
        """
        return next((crate for crate in self.crates if crate.tile_position == position), None)

    def can_move_crate_to(self, position: TilePosition) -> bool:
        """Test whether a crate can be moved to given position.

        Crate can be moved to a position only when there are no obstacles there (crates, walls).

        :param position: position to test whether the crate can be moved to
        :return: True - if crate can be moved to given location *OR* False - otherwise
        """
        return self.template.tile_at(position).is_walkable and not self.crate_at_pos(position)

    def is_crate_in_place(self, position: TilePosition) -> bool:
        """Test whether the crate is "in place" (at cargo bay) in given position.

        :param position: position to test whether the crate is "in place" or not
        :return: True - if crate is "in place" at given position *OR* False - otherwise
        """
        tile = self.template.tile_at(position)
        return tile.is_cargo_bay or tile.is_crate_initially_placed

    def process_input(self, input_action: Optional[InputAction]) -> None:
        """Transform given input action to game action and queue it (so it can be run later,
        during update call)."""

        if self.running_action:
            # TODO: Add movement cancellation when movement with opposite direction was triggered
            #       (only when robot is not pushing a crate)
            return

        # TODO: If player is pressing more then one directional button, prefer the one which
        #       does not lead to collision

        # Before checking the input we don't know whether there will be a continuation of the
        # movement or not. That's why we cannot put robot to standing state when action finishes.
        # We have to do it here. The same applies to last_input_action.
        self.robot.robot_state = RobotState.STANDING
        self.last_input_action = input_action

        if not input_action:
            return

        if input_action == InputAction.UNDO and self.history:
            self.running_action = self.history.pop()
            self.running_action.reset(backward=True)
        if input_action.is_movement:
            robot_dest = self.robot.tile_position.move(input_action.direction)
            if self.template.tile_at(robot_dest).is_walkable:
                crate = self.crate_at_pos(robot_dest)
                if crate:
                    crate_dest = crate.tile_position.move(input_action.direction)
                    if self.can_move_crate_to(crate_dest):
                        self.running_action = PushCrate(self.robot, crate, input_action.direction)
                else:
                    self.running_action = MoveRobot(self.robot, input_action.direction)
            else:
                self.running_action = TurnRobot(self.robot, input_action.direction)

    def update(self, dt_in_ms: float) -> None:
        """Perform an update on the level's game logic."""
        self._update_running_action(dt_in_ms)
        self._evaluate_crates()
        for game_object in self.game_objects:
            game_object.update(dt_in_ms)
        self.statistics.game_time += dt_in_ms

    def draw(self) -> None:
        """Draw all layers of level in order (from bottom to top)."""
        for layer in self.template.layers:
            self.template.tilemap.draw(layer)
            for game_object in self.game_objects:
                game_object.draw(layer)

    def _update_running_action(self, dt_in_ms: float) -> None:
        if self.running_action:
            last_action = self.running_action
            self.running_action = self.running_action.update(dt_in_ms, self.statistics)

            if last_action is not self.running_action and not last_action.backward:
                self.history.append(last_action)

    def _evaluate_crates(self) -> None:
        for crate in self.crates:
            crate_in_place = self.is_crate_in_place(crate.tile_position)
            crate.state = CrateState.PLACED if crate_in_place else CrateState.MISPLACED



"""Module exposing JSON schema for Bansoko's resources metadata file."""

METADATA_JSON_SCHEMA = {
    "$schema": "http://json-schema.org/draft-07/schema#",
    "title": "Bansoko resources metadata schema",
    "description": "Schema describing format of metadata for game resources",
    "type": "object",
    "definitions": {
        "resource_name": {
            "type": "string",
            "description": "Name of a game resource that can be referenced",
            "pattern": "^[A-Za-z_][A-Za-z0-9_]*$"
        },
        "point": {
            "type": "array",
            "description": "A point representing a location on screen",
            "items": {
                "type": "integer"
            },
            "minItems": 2,
            "maxItems": 2
        },
        "rect": {
            "type": "array",
            "description": "A rectangle expressed as position and size",
            "items": {
                "type": "integer"
            },
            "minItems": 4,
            "maxItems": 4
        },
        "color": {
            "type": "integer",
            "description": "One of 16 available colors",
            "minimum": 0,
            "maximum": 15
        },
        "transparency_color": {
            "type": "integer",
            "description": "Transparency color is a color which is ignored during drawing."
                           "One of 16 available colors *OR* -1 for no transparency",
            "minimum": -1,
            "maximum": 15
        },
        "image_bank": {
            "type": "integer",
            "description": "The id of Pyxel's image bank",
            "minimum": 0,
            "maximum": 1
        },
        "tilemap_id": {
            "type": "integer",
            "description": "The id of Pyxel's tilemap",
            "minimum": 0,
            "maximum": 7
        },
    },
    "properties": {
        "sprites": {
            "description": "Collection of sprites that can be used in the game",
            "type": "object",
            "propertyNames": {
                "$ref": "#/definitions/resource_name"
            },
            "additionalProperties": {
                "type": "object",
                "properties": {
                    "image_bank": {
                        "description": "Image bank the sprite is located in",
                        "$ref": "#/definitions/image_bank"
                    },
                    "directional": {
                        "description": "Is sprite directional. Directional sprites has variants "
                                       "for all 4 directions (Left, Right, Up and Down).",
                        "type": "boolean"
                    },
                    "transparency_color": {
                        "description": "Transparency color for the sprite",
                        "$ref": "#/definitions/transparency_color"
                    },
                    "num_frames": {
                        "description": "Number of animation frames in the sprite",
                        "type": "integer",
                        "minimum": 1
                    },
                    "num_layers": {
                        "description": "Number of layers of the sprite",
                        "type": "integer",
                        "minimum": 1
                    },
                    "uv_rect": {
                        "description": "The sprite coordinates in the image bank",
                        "$ref": "#/definitions/rect"
                    }
                },
                "required": ["image_bank", "directional", "transparency_color", "num_frames",
                             "num_layers", "uv_rect"]
            }
        },
        "sprite_packs": {
            "description": "Collection of sprite packs that can be used in the game",
            "type": "object",
            "propertyNames": {
                "$ref": "#/definitions/resource_name"
            },
            "additionalProperties": {
                "type": "array",
                "items": {
                    "$ref": "#/definitions/resource_name"
                },
                "minItems": 1
            }
        },
        "screens": {
            "description": "Collection of screens that can be used in the game",
            "type": "object",
            "propertyNames": {
                "$ref": "#/definitions/resource_name"
            },
            "additionalProperties": {
                "type": "object",
                "properties": {
                    "background": {
                        "type": "object",
                        "description": "",
                        "properties": {
                            "background_color": {
                                "$ref": "#/definitions/color",
                                "description": "Background color screen is cleared with"
                            },
                            "background_tilemap": {
                                "type": "object",
                                "description": "Background tilemap drawn on the screen "
                                               "(it's drawn as first)",
                                "properties": {
                                    "tilemap_id": {
                                        "$ref": "#/definitions/tilemap_id",
                                        "description": "Pyxel's tilemap id for the background "
                                                       "tilemap"
                                    },
                                    "tilemap_uv": {
                                        "$ref": "#/definitions/rect",
                                        "description": "Rectangle representing a portion of "
                                                       "Pyxel's tilemap that is going to be drawn "
                                                       "on game screen"
                                    }
                                },
                                "required": ["tilemap_id", "tilemap_uv"]
                            }
                        },
                        "additionalProperties": False
                    },
                    "elements": {
                        "type": "array",
                        "description": "Collection of elements (images, texts) of the game screen",
                        "items": {
                            "type": "object",
                            "properties": {
                                "position": {
                                    "$ref": "#/definitions/point",
                                    "description": "Position of the element in screen space"
                                },
                                "sprite_ref": {
                                    "$ref": "#/definitions/resource_name",
                                    "description": "Sprite of the element, if element is sprite "
                                                   "based"
                                },
                                "text": {
                                    "type": "string",
                                    "description": "Text of the element, if element is text based"
                                }
                            },
                            "oneOf": [
                                {
                                    "required": ["position", "sprite_ref"]
                                },
                                {
                                    "required": ["position", "text"]
                                }
                            ]
                        }
                    },
                    "menu": {
                        "type": "object",
                        "description": "Configuration of the menu displayed on the game screen",
                        "properties": {
                            "position": {
                                "$ref": "#/definitions/point",
                                "description": "Position of the menu in screen space"
                            },
                            "scrollbar_rect": {
                                "$ref": "#/definitions/rect",
                                "description": "Rectangle describing the position and size of "
                                               "the menu scrollbar"
                            }
                        },
                        "additionalProperties": False
                    }
                }
            }
        },
        "levels": {
            "type": "object",
            "properties": {
                "sha1": {
                    "type": "string",
                    "description": "SHA1 of the bundle metadata file describes. It's generated"
                                   "based on bundle name and levels.",
                    "pattern": "^[a-f0-9]{40}$"
                },
                "level_templates": {
                    "type": "array",
                    "description": "Collection of templates used when instantiating levels",
                    "items": {
                        "type": "object",
                        "properties": {
                            "tileset": {
                                "type": "integer",
                                "description": "The id of tileset level template uses",
                                "minimum": 0
                            },
                            "draw_offset": {
                                "$ref": "#/definitions/point",
                                "description": "The offset of the level's tilemap relative to "
                                               "screen"
                            },
                            "robot_sprite_pack_ref": {
                                "$ref": "#/definitions/resource_name",
                                "description": "Reference to sprite pack containing robot sprites"
                            },
                            "crate_sprite_pack_ref": {
                                "$ref": "#/definitions/resource_name",
                                "description": "Reference to sprite pack containing crate sprites"
                            }
                        }
                    }
                }
            }
        }
    },
    "additionalProperties": False
}
